# LLM constants

B_INST, E_INST = "[INST]", "[/INST]"
B_SYS, E_SYS = "<<SYS>>\n", "\n<</SYS>>\n\n"


class LlmCT:
  P_USER_START = B_INST
  P_USER_END = E_INST
  P_ROUND_START = '<s>'
  P_ROUND_END = '</s>'
  P_SYS_START = B_SYS
  P_SYS_END = E_SYS

  HIST = 'history'
  REQ = 'request'
  RES = 'response'
  SYS = 'system_info'
  CTX = 'context'

  PRED = 'prediction'
  TEXT = 'text'
  TKNS = 'tokens'
  PRMP = 'prompt'
  TPS  = 'tps'

  # Constants for encoding a prompt using chat templates
  REQUEST_ROLE = 'user'
  REPLY_ROLE = 'assistant'
  SYSTEM_ROLE = 'system'
  ROLE_KEY = 'role'
  DATA_KEY = 'content'

  EE_HF_TOKEN = 'EE_HF_TOKEN'

  LLAMA3_CHAT_TEMPLATE = """{{ bos_token }}
{% if messages[0]['role'] == 'system' %}
    {% set loop_messages = messages[1:] %}
    {% set system_message = '<|start_header_id|>' + 'system' + '<|end_header_id|>\n\n' + messages[0]['content'].strip() + '<|eot_id|>' %}
{% else %}
    {% set loop_messages = messages %}
    {% set system_message = '' %}
{% endif %}

{% for message in loop_messages %}
    {% if (message['role'] == 'user') != (loop.index0 % 2 == 0) %}
        {{ raise_exception('Conversation roles must alternate user/assistant/user/assistant/...') }}
    {% endif %}

    {% if loop.index0 == 0 %}
        {{ system_message }}
    {% endif %}

    {{ '<|start_header_id|>' + message['role'] + '<|end_header_id|>\n\n' + message['content'].strip() + '<|eot_id|>' }}

    {% if loop.last and message['role'] == 'user' and add_generation_prompt %}
        {{ '<|start_header_id|>' + 'assistant' + '<|end_header_id|>\n\n' }}
    {% endif %}
{% endfor %}
"""

  MISTRAL_CHAT_TEMPLATE = """{% if messages[0]['role'] == 'system' %}
    {% set loop_messages = messages[1:] %}
    {% set system_message = messages[0]['content'].strip() + '\n\n' %}
{% else %}
    {% set loop_messages = messages %}
    {% set system_message = '' %}
{% endif %}

{{ bos_token }}
{% for message in loop_messages %}
    {% if (message['role'] == 'user') != (loop.index0 % 2 == 0) %}
        {{ raise_exception('Conversation roles must alternate user/assistant/user/assistant/...') }}
    {% endif %}

    {% if loop.index0 == 0 %}
        {% set content = system_message + message['content'] %}
    {% else %}
        {% set content = message['content'] %}
    {% endif %}

    {% if message['role'] == 'user' %}
        {{ '[INST] ' + content.strip() + ' [/INST]' }}
    {% elif message['role'] == 'assistant' %}
        {{ ' ' + content.strip() + eos_token }}
    {% endif %}
{% endfor %}
"""

# END LLM constants




from naeural_core.business.default.web_app.fast_api_web_app import FastApiWebAppPlugin as BasePlugin

__VER__ = '0.1.0.0'

_CONFIG = {
  **BasePlugin.CONFIG,
  
  "RUNS_ONLY_ON_SUPERVISOR_NODE" : True,

  'VALIDATION_RULES': {
    **BasePlugin.CONFIG['VALIDATION_RULES'],
  },
}


class SupervisorFastApiWebApp(BasePlugin):
  CONFIG = _CONFIG

  def on_init(self):
    self.__supervisor_fastapi_plugin_running = None
    super(SupervisorFastApiWebApp, self).on_init()
    return

  @property
  def __is_enabled(self):
    res = not self.cfg_disabled and self.cfg_ngrok_edge_label is not None and self.is_supervisor_node
    if res != self.__supervisor_fastapi_plugin_running:
      self.__supervisor_fastapi_plugin_running = res
      if res:
        self.P(f"{self.__class__.__name__} is enabled")
      else:
        disabled_cause = 'disabled by config parameter' if self.cfg_disabled \
          else 'no ngrok edge label provided' if self.cfg_ngrok_edge_label is None \
          else 'not a supervisor node'
        msg = f"{self.__class__.__name__} is disabled. (cause: {disabled_cause})"
        self.P(msg, color='r', boxed=True)
    # endif changed state
    return res

  def _process(self):
    if not self.__is_enabled:
      return None
    return super(SupervisorFastApiWebApp, self)._process()




class _BasePluginAPIMixin:
  def __init__(self) -> None:
    super(_BasePluginAPIMixin, self).__init__()
    
    self.__chain_state_initialized = False
    return
  
  # Obsolete
  def _pre_process(self):
    """
    Called before process. Currently (partially) obsolete

    Returns
    -------
    TBD.

    """
    return
  
  def _post_process(self):
    """
    Called after process. Currently (partially) obsolete

    Returns
    -------
    TBD.

    """
    return
  
  
  def step(self):
    """
    The main code of the plugin (loop iteration code). Called at each iteration of the plugin loop.

    Returns
    -------
    None.

    """
    return
  
  
  def process(self):
    """
    The main code of the plugin (loop iteration code). Called at each iteration of the plugin loop.

    Returns
    -------
    Payload.

    """
    return self.step()
  
  def _process(self):
    """
    The main code of the plugin (loop iteration code.

    Returns
    -------
    Payload.

    """
    return self.process()

  
  def on_init(self):
    """
    Called at init time in the plugin thread.

    Returns
    -------
    None.

    """      
    return
  
  def _on_init(self):
    """
    Called at init time in the plugin thread.

    Returns
    -------
    None.

    """
    self.P("Default plugin `_on_init` called for plugin initialization...")
    self.on_init()
    return


  def on_close(self):
    """
    Called at shutdown time in the plugin thread.

    Returns
    -------
    None.

    """      
    return


  def _on_close(self):
    """
    Called at shutdown time in the plugin thread.

    Returns
    -------
    None.

    """
    self.P("Default plugin `_on_close` called for plugin cleanup at shutdown...")
    self.maybe_archive_upload_last_files()
    self.on_close()
    return

  def on_command(self, data, **kwargs):
    """
    Called when the instance receives new INSTANCE_COMMAND

    Parameters
    ----------
    data : any
      object, string, etc.

    Returns
    -------
    None.

    """
    return

  def _on_command(self, data, default_configuration=None, current_configuration=None, **kwargs):
    """
    Called when the instance receives new INSTANCE_COMMAND

    Parameters
    ----------
    data : any
      object, string, etc.

    Returns
    -------
    None.

    """
    self.P("Default plugin `_on_command`...")

    if (isinstance(data, str) and data.upper() == 'DEFAULT_CONFIGURATION') or default_configuration:
      self.P("Received \"DEFAULT_CONFIGURATION\" command...")
      self.add_payload_by_fields(
        default_configuration=self._default_config,
        command_params=data,
      )
      return
    if (isinstance(data, str) and data.upper() == 'CURRENT_CONFIGURATION') or current_configuration:
      self.P("Received \"CURRENT_CONFIGURATION\" command...")
      self.add_payload_by_fields(
        current_configuration=self._upstream_config,
        command_params=data,
      )
      return

    self.on_command(data, **kwargs)
    return


  def _on_config(self):
    """
    Called when the instance has just been reconfigured

    Parameters
    ----------
    None

    Returns
    -------
    None.

    """
    self.P("Default plugin {} `_on_config` called...".format(self.__class__.__name__))
    if hasattr(self, 'on_config'):
      self.on_config()
    return


  ###
  ### Chain State
  ### 
  
  def __maybe_wait_for_chain_state_init(self):
    # TODO: raise exception if not found after a while

    while not self.plugins_shmem.get('__chain_storage_set'):
      self.sleep(0.1)
    
    if not self.__chain_state_initialized:
      self.P(" ==== Chain state initialized.")
    self.__chain_state_initialized = True
    return
  
  def chainstore_set(self, key, value, debug=False):
    result = False
    try:
      self.__maybe_wait_for_chain_state_init()
      func = self.plugins_shmem.get('__chain_storage_set')
      if func is not None:
        if debug:
          self.P("Setting data: {} -> {}".format(key, value), color="green")
        self.start_timer("chainstore_set")
        result = func(key, value, debug=debug)
        elapsed = self.end_timer("chainstore_set")        
        if debug:
          self.P(" ====> `chainstore_set` elapsed time: {:.6f}".format(elapsed), color="green")
      else:
        if debug:
          self.P("No chain storage set function found", color="red")
    except Exception as ex:
      self.P("Error in chainstore_set: {}".format(ex), color="red")
    return result
  
  
  def chainstore_get(self, key, debug=False):
    self.__maybe_wait_for_chain_state_init()
    value = None
    try:
      while self.plugins_shmem.get('__chain_storage_get') is None:
        self.sleep(0.1)
        # TODO: raise exception if not found after a while
      func = self.plugins_shmem.get('__chain_storage_get')
      if func is not None:
        value = func(key, debug=debug)
        if debug:
          self.P("Getting data: {} -> {}".format(key, value), color="green")
      else:
        if debug:
          self.P("No chain storage get function found", color="red")
    except Exception as ex:
      self.P("Error in chainstore_get: {}".format(ex), color="red")
    return value
  
  
  # # @property
  # # This CANNOT be a property, as it can be a blocking operation.
  # def _chainstorage(self): # TODO: hide/move/protect this
  #   self.__maybe_wait_for_chain_state_init()
  #   return self.plugins_shmem.get('__chain_storage')

  
  def get_instance_path(self):
    return [self.ee_id, self._stream_id, self._signature, self.cfg_instance_id]  
  
  ###
  ### END Chain State
  ###
  
    


import torch as th

from transformers import AutoTokenizer as LlmTokenizer
from transformers import AutoModelForCausalLM as LlmForCausalLM
from transformers import BitsAndBytesConfig


class LlmModelMixin(object):
  def __init__(self, *args, **kwargs):
    super(LlmModelMixin, self).__init__(*args, **kwargs)
    return


  def _get_placement_summary(self, indent=4):
    """Logs the device map from the model.

    Parameters
    ----------
      None.
    """

    def str_device(dev):
      str_place = dev
      str_place = str_place if str_place in ['cpu', 'disk'] else 'cuda:' + str_place
      return str_place

    str_indent = " " * indent
    result = ""
    if hasattr(self.model, 'hf_device_map'):
      self.placement = self.model.hf_device_map
      device = None
      prev_layer = None
      n = 0
      if len(self.placement) == 1:
        _layer = list(self.placement.keys())[0]
        result = str(self.placement[_layer])
      else:
        for layer in self.placement:
          if device != self.placement[layer]:
            if device is not None:
              result = result + prev_layer + ']({} layers): {}\n'.format(n, str_device(self.placement[layer]))
              n = 0
            device = self.placement[layer]
            result = result + str_indent + '[{} to '.format(layer)
            prev_layer = layer
          n += 1
        result = result + layer + ']({} layers): {}\n'.format(n, str_device(self.placement[layer]))
    return result


  def _get_model_load_config(self):
    return self.log.get_model_load_config(
      model_name=self.cfg_model_name,
      token=self.hf_token,
      has_gpu=self.has_gpu,
      weights_size=self.cfg_model_weights_size,
      device_map=self._get_device_map(),
      cache_dir=self.cache_dir,
      use_flash_attention=self.cfg_use_flash_attention
    )

  def load_tokenizer(self, model_id, cache_dir, token):
    """
    Load the tokenizer from the model and set up padding.
    Parameters
    ----------
    model_id : str
        the model identifier
    cache_dir : str
        the cache directory
    token : str
        the token to use for authentication
    Returns
    -------

    """
    self.tokenizer = LlmTokenizer.from_pretrained(
      model_id,
      cache_dir=cache_dir,
      use_auth_token=token,
    )

    # Fix for missing system roles in transformers.
    self._set_tokenizer_chat_template()
    return

  def _load_tokenizer(self):
    """Loads the tokenizer from the model and sets up padding.
    """
    # Load the tokenizer and output to log.
    cache_dir = self.cache_dir
    token = self.hf_token
    model_id = self.cfg_model_name
    self.P("Loading tokenizer for {} in '{}'...".format(model_id, cache_dir))
    self.load_tokenizer(model_id, cache_dir, token)

    # Use the unknown token as the padding token. It seems that at least
    # when quantized llama2 will look at the embeddings of padding tokens
    # so we should use something that is as ignorable as possible
    # embedding-wise.
    self.tokenizer.padding_side = 'right'
    if self.tokenizer.pad_token is None:
      self.tokenizer.pad_token = self.tokenizer.unk_token
    if self.tokenizer.pad_token is None:
      self.tokenizer.pad_token = self.tokenizer.eos_token

    self.padding_id = self.tokenizer.pad_token_id
    if self.padding_id is None:
      self.padding_id = self.tokenizer.unk_token_id
    if self.padding_id is None:
      self.padding_id = self.tokenizer.eos_token_id
    self.P(
      'Settting padding token to {} and padding token id to {}'
      .format(
        self.tokenizer.pad_token, self.tokenizer.pad_token_id
      )
    )

    self.P("  Loaded `{}` tokenizer".format(self.tokenizer.__class__.__name__))
    return

  def load_pretrained_model(self, model_id, **kwargs):
    """
    Load the pretrained model with the given model id and additional parameters.
    Parameters
    ----------
    model_id  : str - the model identifier
    kwargs : dict - additional parameters

    Returns
    -------
    model : _BaseAutoModelClass - the loaded model
    """
    return LlmForCausalLM.from_pretrained(model_id, **kwargs)

  def _load_model(self):
    """
    Load the model from the given configured model name and set up padding.
    Will first set up the model loading configuration and then load the model

    """
    model_id = self.cfg_model_name
    model_params, quantization_params = self._get_model_load_config()
    self.P("Loading {} with following parameters:\n{}\nQuantization params: {}".format(
      model_id,
      self.json_dumps(model_params, indent=4),
      self.json_dumps(quantization_params, indent=4),
      )
    )

    quantization_config = None
    if quantization_params is not None:
      quantization_config = BitsAndBytesConfig(**quantization_params)
    model_params['quantization_config'] = quantization_config

    self.P(f'Trying to load pretrained for {model_id} with the following params:\n {model_params}')

    self.model = self.load_pretrained_model(model_id, **model_params)

    compiled = self.cfg_th_compile
    if compiled:
      compile_mode = self.cfg_th_compile_mode
      self.P("Compiling model")
      self.model = th.compile(
        self.model,
        fullgraph=True,
        mode=compile_mode
      )
    #endif compile model

    self.P("  Loaded `{}` model".format(self.model.__class__.__name__))

    # Set the padding token to the chosen (<unk>) token.
    self.model.config.pad_token_id = self.padding_id
    self.P('Setting padding token ID to {}'.format(self.model.config.pad_token_id))

    # When the entire model is on the GPU we expect to get a {'':0} device map
    # which doesn't really tell us where the model is, only that it is on one
    # device. Additionally print the model device to avoid this corner case.
    self.P("Model {} loaded with dev map:\n{}".format(model_id, self._get_placement_summary()))
    device = next(self.model.parameters()).device
    self.P("First weight is on device: {}".format(device))

    return



from naeural_core.bc import DefaultBlockEngine
class BCWrapper:
  def __init__(self, blockchain_manager : DefaultBlockEngine):
    self.__bc : DefaultBlockEngine = blockchain_manager
    return
  
  @property
  def address(self):
    """
    Returns the address of the current node

    Returns
    -------
    str
        The address of the current node in the blockchain
    """
    return self.__bc.address
  
  def sign(
    self, 
    dct_data: dict, 
    add_data: bool = True, 
    use_digest: bool = True, 
    replace_nan: bool = True
  ) -> str:
    """
    Generates the signature for a dict object.
    Does not add the signature to the dict object

    Parameters
    ----------
    dct_data : dict
      the input message as a dict.
      
    add_data: bool, optional
      will add signature and address to the data dict (also digest if required). Default `True`
      
    use_digest: bool, optional  
      will compute data hash and sign only on hash
      
    replace_nan: bool, optional
      will replace `np.nan` and `np.inf` with `None` before signing. 

    Returns
    -------
      text signature

        
      IMPORTANT: 
        It is quite probable that the same sign(sk, hash) will generate different signatures
    """
    return self.__bc.sign(
      dct_data=dct_data,
      add_data=add_data,
      use_digest=use_digest,
      replace_nan=replace_nan
    )
  
  def verify(
    self, 
    dct_data: str, 
    str_signature: str = None, 
    sender_address: str = None,
    return_full_info : bool = True,
  ) -> bool:
    """
    Verifies a signature using the public key of the signer

    Parameters
    ----------
    dct_data : dict
        the data that was signed
        
    str_signature : str
        the base64 encoded signature. Default `None` will be taken from the data
        
    str_signer : str
        the signer's address (string) used as the public key for verification.
        Default `None` will be taken from the data
        
    return_full_info: bool, optional
        whether to return the full verification info. Default `True`      
    

    Returns
    -------
    bool
        True if the signature is valid, False otherwise
    """
    return self.__bc.verify(
      dct_data=dct_data, 
      signature=str_signature, 
      sender_address=sender_address,
      return_full_info=return_full_info,
    )

  def encrypt_str(self, str_data : str, str_recipient : str):
    """
    Encrypts a string using the public key of the recipient using asymmetric encryption

    Parameters
    ----------
    str_data : str
        the data to be encrypted (string)
        
    str_recipient : str
        the recipient's address (string) used as the public key
        
    OBSOLETE:
      compress: bool, optional
          whether to compress the data before encryption. Default `True`

    Returns
    -------
    str
       the base64 encoded encrypted data
    """
    encrypted_data = self.__bc.encrypt(
      plaintext=str_data, receiver_address=str_recipient,
    )
    return encrypted_data
  
  def decrypt_str(self, str_b64data : str, str_sender : str):
    """
    Decrypts a base64 encoded string using the private key of the sender using asymmetric encryption.
    This method is able to decrypt data that was was encrypted for multiple recipients as well.

    Parameters
    ----------
    str_b64data : str
        The base64 encoded encrypted data
        
    str_sender : str
        The sender's address (string) used as the public key for decryption
        
    OBSOLETE:
      embed_compressed: bool, optional
          whether the compression flag is embedded in the data. Default `True`. Modify this only for special cases.

    Returns
    -------
    str
       the decrypted data (string) that can be then decoded to the original data
    """
    decompressed_data = self.__bc.decrypt(
      encrypted_data_b64=str_b64data, sender_address=str_sender,
    )
    return decompressed_data
  
  
  def encrypt_multi(self, str_data : str, lst_recipients : list):
    """
    Encrypts a string using the public key of the recipient using asymmetric encryption

    Parameters
    ----------
    str_data : str
        the data to be encrypted (string)
        
    lst_recipients : list
        the list of recipient's addresses (string) used as the public keys

    Returns
    -------
    str
       the base64 encoded encrypted data
    """
    encrypted_data = self.__bc.encrypt_for_multi(
      plaintext=str_data, receiver_addresses=lst_recipients,
    )
    return encrypted_data


  def get_whitelist(self, with_prefix: bool = False):
    """
    Returns the list of nodes that are allowed to connect to the current node

    Returns
    -------
    list
        The list of addresses that are whitelisted
    """
    lst_allowed = self.__bc.whitelist
    if with_prefix:
      lst_allowed = [self.maybe_add_prefix(addr) for addr in lst_allowed]
    return lst_allowed


  def get_allowed_nodes(self, with_prefix: bool = False):
    """
    Returns the list of nodes that are allowed to connect to the current node. Alias of `get_whitelist`

    Returns
    -------
    list
        The list of addresses that are allowed to connect
    """
    return self.get_whitelist(with_prefix=with_prefix)


  def maybe_remove_addr_prefix(self, address: str):
    """
    Removes the address prefix from the current node's address
    
    Parameters
    ----------
    
    address: str
        The address to remove the prefix from

    Returns
    -------
    str
        The address of the current node without the prefix
    """
    return self.__bc.maybe_remove_prefix(address)
  
  
  def maybe_remove_prefix(self, address: str):
    """
    Removes the prefix from the address. Alias of `maybe_remove_addr_prefix`
    
    Parameters
    ----------
    
    address: str
        The address to remove the prefix from

    Returns
    -------
    str
        The address without the prefix
    """
    return self.maybe_remove_addr_prefix(address)
  
  
  def maybe_add_prefix(self, address: str):
    """
    Adds the prefix to the address
    
    Parameters
    ----------
    
    address: str
        The address to add the prefix to

    Returns
    -------
    str
        The address with the prefix
    """
    return self.__bc._add_prefix(address)
  


### EVM


  @property
  def eth_address(self):
    """
    Returns the EVM address of the current node. 

    Returns
    -------
    str
        The address of the current node in the blockchain
    """
    return self.__bc.eth_address
  

  def eth_sign_message(self, types: list, values: list):
    """
    Signs a message using the EVM account of the current node

    Parameters
    ----------
    types : list
        The types of the values to be signed
        
    values : list
        The values to be signed

    Returns
    -------
    dict
        A dictionary with the following keys
        - types: list
            The types of the values that were signed
        - values: list
            The values that were signed
        - signature: str
            The signature of the data
        - address: str
            The address of the EVM account used to sign the data
    """
    return self.__bc.eth_sign_message(types=types, values=values)  
  
  
  def eth_sign_node_epochs(self, node: str, epochs: list, epochs_vals: list, signature_only: bool = True):
    """
    Signs the epochs availabilities for a given node using the EVM account of the current node

    Parameters
    ----------
    node : str
        The address of the node
        
    epochs : list
        The epochs to be signed
        
    epochs_vals : list
        The values of the epochs to be signed
        
    signature_only : bool, optional
        Whether to return only the signature. Default `False`

    Returns
    -------
    dict or just str
        A dictionary with the following keys
        - node: str
            The address of the node
        - epochs_vals: list
            The values of the epochs that were signed
        - eth_signature: str
            The EVM signature of the data
        - eth_address: str
            The address of the EVM account used to sign the data
    """
    return self.__bc.eth_sign_node_epochs(
      node=node, epochs=epochs, epochs_vals=epochs_vals, signature_only=signature_only
    )
    
    
  def node_address_to_eth_address(self, node_address: str):
    """
    Converts a node address to an EVM address

    Parameters
    ----------
    node_address : str
        The address of the node

    Returns
    -------
    str
        The EVM address of the node
    """
    return self.__bc.node_address_to_eth_address(node_address)
  


from naeural_core.serving.base.base_doc_emb_serving import BaseDocEmbServing as BaseServingProcess

__VER__ = '0.1.0.0'

_CONFIG = {
  **BaseServingProcess.CONFIG,

  "MODEL_NAME": "mixedbread-ai/mxbai-embed-large-v1",

  'VALIDATION_RULES': {
    **BaseServingProcess.CONFIG['VALIDATION_RULES'],
  },

}


class MxbaiEmbed(BaseServingProcess):
  CONFIG = _CONFIG



"""
```bibtex
@misc{touvron2023llama,
      title={Llama 2: Open Foundation and Fine-Tuned Chat Models},
      author={Hugo Touvron and Louis Martin and Kevin Stone and Peter Albert and Amjad Almahairi and Yasmine Babaei and Nikolay Bashlykov and Soumya Batra and Prajjwal Bhargava and Shruti Bhosale and Dan Bikel and Lukas Blecher and Cristian Canton Ferrer and Moya Chen and Guillem Cucurull and David Esiobu and Jude Fernandes and Jeremy Fu and Wenyin Fu and Brian Fuller and Cynthia Gao and Vedanuj Goswami and Naman Goyal and Anthony Hartshorn and Saghar Hosseini and Rui Hou and Hakan Inan and Marcin Kardas and Viktor Kerkez and Madian Khabsa and Isabel Kloumann and Artem Korenev and Punit Singh Koura and Marie-Anne Lachaux and Thibaut Lavril and Jenya Lee and Diana Liskovich and Yinghai Lu and Yuning Mao and Xavier Martinet and Todor Mihaylov and Pushkar Mishra and Igor Molybog and Yixin Nie and Andrew Poulton and Jeremy Reizenstein and Rashi Rungta and Kalyan Saladi and Alan Schelten and Ruan Silva and Eric Michael Smith and Ranjan Subramanian and Xiaoqing Ellen Tan and Binh Tang and Ross Taylor and Adina Williams and Jian Xiang Kuan and Puxin Xu and Zheng Yan and Iliyan Zarov and Yuchen Zhang and Angela Fan and Melanie Kambadur and Sharan Narang and Aurelien Rodriguez and Robert Stojnic and Sergey Edunov and Thomas Scialom},
      year={2023},
      eprint={2307.09288},
      archivePrefix={arXiv},
      primaryClass={cs.CL}
}
```

```bibtex
@misc{rozière2023code,
      title={Code Llama: Open Foundation Models for Code},
      author={Baptiste Rozière and Jonas Gehring and Fabian Gloeckle and Sten Sootla and Itai Gat and Xiaoqing Ellen Tan and Yossi Adi and Jingyu Liu and Tal Remez and Jérémy Rapin and Artyom Kozhevnikov and Ivan Evtimov and Joanna Bitton and Manish Bhatt and Cristian Canton Ferrer and Aaron Grattafiori and Wenhan Xiong and Alexandre Défossez and Jade Copet and Faisal Azhar and Hugo Touvron and Louis Martin and Nicolas Usunier and Thomas Scialom and Gabriel Synnaeve},
      year={2023},
      eprint={2308.12950},
      archivePrefix={arXiv},
      primaryClass={cs.CL}
}
```

The inputs of the plugin must be in the following format for simplest payload:
```json
{
  "ACTION" : "PIPELINE_COMMAND",
  "PAYLOAD" :
  {
    "NAME": "llm-on-demand",
    "PIPELINE_COMMAND" :
    {
      "STRUCT_DATA" :
      {
        "request" : "write a hello world program in C++"
      }
    }
  }
}
```

and for history based command:
```json
{
  "ACTION" : "PIPELINE_COMMAND",
  "PAYLOAD" : {
    "NAME": "llm-on-demand",
    "PIPELINE_COMMAND" : {
      "STRUCT_DATA" : [{
        "request" : "return ",
        "history" : [
          {
            "request"   : "print('hello",
            "response"  : " world')"
          }
        ],
        "system_info" : "You are a funny python programmer assistant. your task is to complete the code you are given. return only the completion, not the whole program."
      }]
    }
  }
}
```



"""
import gc
import numpy as np
import torch as th
import transformers
import tokenizers
import accelerate



from naeural_core.serving.mixins_llm import LlmTokenizerMixin
from naeural_core.serving.mixins_llm import LlmModelMixin
from naeural_core.serving.mixins_llm.llm_utils import LlmCT

from naeural_core.serving.base import ModelServingProcess as BaseServingProcess

TEST_MODULES = [
  th,
  transformers,
  tokenizers,
  accelerate
]


__VER__ = '0.1.0.2'

_CONFIG = {
  **BaseServingProcess.CONFIG,

  "DEFAULT_DEVICE"        : "cuda:0",

  "MAX_WAIT_TIME"         : 1000,
  "SERVER_COLLECTOR_TIMEDELTA": 172800,  # 48 hours -> this is done because the llm model is very large and
  # we want to keep it in memory for a long time.

  "PICKED_INPUT"          : "STRUCT_DATA",

  "RUNS_ON_EMPTY_INPUT"   : False,

  "MODEL_NAME"            : None,

  "REPETITION_PENALTY"    : 1.1,

  "ADD_SPECIAL_TOKENS": False,
  "ADD_GENERATION_PROMPT": False,

  "TH_COMPILE"            : False,

  "TH_COMPILE_MODE"       : "max-autotune",

  "USE_FLASH_ATTENTION"   : False,

  # Possible values of None, 4, 8, 16, 32
  # where None is the default model config.
  "MODEL_WEIGHTS_SIZE"    : None,

  # Number of tokens overlapping when decoding. Used for Prompt lookup decoding.
  # If None, the model will not use Prompt lookup decoding.
  "PROMPT_LOOKUP_NUM_TOKENS": None,
  # To be adjusted in the future
  "HISTORY_LIMIT": 20,

  'VALIDATION_RULES': {
    **BaseServingProcess.CONFIG['VALIDATION_RULES'],
  },

}


class BaseLlmServing(
  BaseServingProcess,
  LlmTokenizerMixin,
  LlmModelMixin,
):
  CONFIG = _CONFIG

  def __init__(self, **kwargs):
    self._counter = 0
    self._version_base = __VER__
    self.model = None
    self.tokenizer = None
    self.device = None
    self.__tps = self.deque(maxlen=128)
    self.padding_id = None
    super(BaseLlmServing, self).__init__(**kwargs)
    return

  @property
  def th(self):
    """
    Proxy to the torch module.
    Returns
    -------
    torch module
    """
    return th

  @property
  def hf_token(self):
    return self.os_environ.get(LlmCT.EE_HF_TOKEN, None)


  @property
  def hf_model(self):
    return self.cfg_model_name


  @property
  def cache_dir(self):
    return self.log.get_models_folder()


  @property
  def has_gpu(self):
    return 'cuda' in self.device.type


  def get_local_path(self):
    models_cache = self.log.get_models_folder()
    model_name = 'models/{}'.format(self.cfg_model_name)
    model_subfolder = model_name.replace('/', '--')
    path = self.os_path.join(models_cache, model_subfolder)
    if self.os_path.isdir(path):
      return path
    else:
      return None

  def get_model_disk_size(self):
    path = self.get_local_path()
    if path is None:
      return 0
    else:
      return self.log.get_folder_size(path)[0]


  def _startup(self):
    # check some params that can be re-configured from biz plugins or
    # (lower priority) serving env in config_startup.txt.
    self.P("Preparing LLM serving...")
    if self.hf_token is None:
      self.P("  No HuggingFace token found. Please set it in the environment variable '{}'".format(LlmCT.EE_HF_TOKEN), color='r')
    else:
      obfuscated = self.hf_token[:3] + '*' * (len(self.hf_token) - 6) + self.hf_token[-3:]
      self.P("  Found HuggingFace token '{}'".format(obfuscated))
    #endif no token

    if self.cfg_model_name is None:
      msg = "No model name found. Please set it in config `MODEL_NAME`"
      raise ValueError(msg)
    #endif no model name

    self.P("  Package versions:")
    for module_test in TEST_MODULES:
      self.P("    {} version: {}".format(module_test.__name__, module_test.__version__))
    #endfor each module

    # setup device
    self._setup_device()

    # setup llm tokenizer
    self._load_tokenizer()

    # setup llm model
    self._load_model()

    # specific llama setup
    self._setup_llm()

    # warm up model
    self._warmup()

    return


  def _warmup(self):
    warmup_request = {
      LlmCT.REQ: "hello",
      LlmCT.HIST: [],
      LlmCT.SYS: "You are a python assistant. Generate some python code."
    }
    # Perform a prediction with a batch of one request.
    warmup_inputs_one = {
      "DATA" : [
        warmup_request
      ]
    }
    # TODO: maybe add warmup flag to predict for printing only in case of warmup
    self._predict(self._pre_process(warmup_inputs_one))
    # Perform a prediction with a batch of four requests.
    warmup_inputs_four = {
      "DATA": [
        warmup_request,
        warmup_request,
        warmup_request,
        warmup_request
      ]
    }
    self._predict(self._pre_process(warmup_inputs_four))
    # Maybe include post_process in the warmup to also check
    # the decoding process.
    self.P("LLM finished warmup")

    return


  def _setup_llm(self):
    raise NotImplementedError("Must be implemented in derived class")
    return


  def _setup_device(self):
    # check if GPU is available & log
    gpu_info = self.log.gpu_info()
    if len(gpu_info) == 0:
      self.device = th.device('cpu')
    else:
      # try default device
      # TODO: review method
      self.device = th.device(self.cfg_default_device)
      device_id = self.device.index
      gpu_name = self.log.get_gpu_name(device_id)
      total_mem = self.log.get_gpu_total_mem(device_id)
      free_mem = self.log.get_gpu_free_mem(device_id)
      self.P("Using default device: {}".format(self.device))
      self.P("  GPU Name:      {}".format(gpu_name))
      self.P("  GPU Total mem: {}".format(total_mem))
      self.P("  GPU Free mem:  {}".format(free_mem))

      disk_size = self.get_model_disk_size()
      self.P("  Model size:    {}".format(disk_size))
      if disk_size > free_mem:
        msg = "  => At default 16bit load model will exceed available GPU memory. Caution is adviced."
        self.P(msg, color='r')
      else:
        msg = "  => At default 16bit load model will fit in GPU memory."
        self.P(msg, color='g')
    return


  def _get_device_map(self):
    # TODO: Rewrite to fix for multiple GPUs
    device_map = "auto"
    return device_map


  def _pre_process(self, inputs):
    lst_inputs = inputs.get('DATA', [])
    serving_params = inputs.get('SERVING_PARAMS', [])
    if True and len(serving_params) > 0:
      self.P("Received full inputs:\n{}".format(self.json_dumps(self.shorten_str(inputs))))
      self.P("Detected 'SERVING_PARAMS': {}".format(serving_params))
    #endif debug
    tokens_lst = []
    predict_kwargs_lst = []
    prompt_lst = []

    for i, inp in enumerate(lst_inputs):
      if not isinstance(inp, dict):
        msg = "Each input must be a dict. Received {}: {}".format(type(inp), self.shorten_str(inputs))
        raise ValueError(msg)
      predict_kwargs = serving_params[i] if i < len(serving_params) else {}
      request = inp.get(LlmCT.REQ, None)
      history = inp.get(LlmCT.HIST, None)
      system_info = inp.get(LlmCT.SYS, None)
      request_context = inp.get(LlmCT.CTX, None)
      prompt = self._get_prompt_from_template(
        request=request,
        history=history,
        system_info=system_info,
        context=request_context
      )
      # Note that we are passing 'pt' in return_tensors to get torch tensors.
      tokens = self.tokenizer.encode(
        prompt,
        add_special_tokens=self.cfg_add_special_tokens,  # False for the majority,
        # Otherwise we would get and extra <bos> at the start.
        # In the case of the pansophic Llama3.1 romanian fine-tuned model, this needs to be True.
        return_tensors='pt'
      ).to(self.device)

      tokens_lst.append(tokens)
      predict_kwargs_lst.append(predict_kwargs)
      prompt_lst.append(prompt)
    #endfor lst_inputs

    # Build the batch tensor. Ideally we should be calling encode on the
    # list of strings directly, however that seems to failing. Additionally
    # __call__ doesn't actually do the infilling.
    max_tok_len = max([toks.shape[1] for toks in tokens_lst])
    batch_tokens = th.ones((len(tokens_lst), max_tok_len), dtype=th.int64, device=self.device) * self.padding_id
    attn_mask = th.zeros((len(tokens_lst), max_tok_len), dtype=th.int64, device=self.device)
    for i, toks in enumerate(tokens_lst):
      batch_tokens[i,:toks.shape[1]] = toks
      attn_mask[i,:toks.shape[1]] = 1

    self.P(f"Generated tokens batch of shape {batch_tokens.shape}")
    self.P(f"Found attention mask of shape {attn_mask.shape}")
    return [batch_tokens, attn_mask, predict_kwargs_lst, prompt_lst]


  def _predict(self, preprocessed_batch):
    self._counter += 1
    batch_tokens, attn_mask, predict_kwargs_lst, prompt_lst = preprocessed_batch
    # Perform generation using tokens and parameters.
    # Note that it's not appropriate to call the forward function
    # here unless we want to re-implement the wheel (various searching
    # strategies i.e. beam searching etc).
    # TODO: change this to pipeline as it seems is the preferred way.
    # TODO: explore more generation strategies, as this is currently
    # using the greedy strategy.

    model_args = {
      'attention_mask': attn_mask,
    }
    if self.cfg_prompt_lookup_num_tokens is not None:
      model_args['prompt_lookup_num_tokens'] = int(self.cfg_prompt_lookup_num_tokens)

    self.P("Running with repetition penalty {}".format(self.cfg_repetition_penalty))

    # TODO: test if some gpu mem can be freed after this
    with th.no_grad():
      t0 = self.time()
      # Note that there's no need to set the padding ID since we've passed
      # the appropriate attention mask.
      # TODO: maybe explore assistant_model parameter from
      #  https://huggingface.co/docs/transformers/v4.44.2/en/llm_optims
      yhat = self.model.generate(
        inputs=batch_tokens,
        max_new_tokens=512,
        repetition_penalty=self.cfg_repetition_penalty,
        **model_args
      )
      elapsed = self.time() - t0
    # endwith
    self.P(f'Done inference in {elapsed} seconds')
    yhat = yhat.cpu().numpy()
    batch_tokens = batch_tokens.cpu().numpy()
    self.th_utils.clear_cache()
    # Calculate number of generated token per seconds and add it to __tps
    # in order to track inference performance. Generated padding is not
    # counted since it is an artefact of the batching strategy.
    batch_y_size = batch_tokens.shape[1]
    num_generated_toks = (yhat[:, batch_y_size:] != self.padding_id).astype(self.np.int32).sum().item()
    num_tps = num_generated_toks / elapsed
    self.__tps.append(num_tps)

    self.P("Model ran at {} tokens per second".format(num_tps))

    dct_result = {
      LlmCT.PRED: yhat,
      LlmCT.PRMP: prompt_lst,
      LlmCT.TKNS: batch_tokens,
      LlmCT.TPS: num_tps,
    }
    return dct_result


  def _post_process(self, preds_batch):
    result = []
    yhat = preds_batch[LlmCT.PRED]
    prompts = preds_batch[LlmCT.PRMP]
    tokens = preds_batch[LlmCT.TKNS]
    tps = preds_batch[LlmCT.TPS]

    # Decode each output in the batch, omitting the input tokens.
    text_lst = self.tokenizer.batch_decode(
      yhat[:,tokens.shape[1]:],
      skip_special_tokens=True
    )

    self.P(f"Found batch text prediction for {len(text_lst)} texts:\n{self.shorten_str(text_lst)}")
    for i, decoded in enumerate(text_lst):
      dct_result = {
        LlmCT.PRED : yhat[i].tolist(),
        LlmCT.PRMP : prompts[i],
        LlmCT.TEXT : decoded,
        LlmCT.TKNS : tokens[i].tolist(),
        LlmCT.TPS  : tps,
        # TODO: find a way to send the model metadata to the plugin, other than through the inferences.
        'MODEL_NAME': self.cfg_model_name
      }
      result.append(dct_result)
    return result




"""
@misc{touvron2023llamaopenefficientfoundation,
      title={LLaMA: Open and Efficient Foundation Language Models},
      author={Hugo Touvron and Thibaut Lavril and Gautier Izacard and Xavier Martinet and Marie-Anne Lachaux and Timothée Lacroix and Baptiste Rozière and Naman Goyal and Eric Hambro and Faisal Azhar and Aurelien Rodriguez and Armand Joulin and Edouard Grave and Guillaume Lample},
      year={2023},
      eprint={2302.13971},
      archivePrefix={arXiv},
      primaryClass={cs.CL},
      url={https://arxiv.org/abs/2302.13971},
}

models:
  meta-llama/Meta-Llama-3.1-8B
  meta-llama/Meta-Llama-3.1-8B-Instruct

  meta-llama/Meta-Llama-3.1-70B
  meta-llama/Meta-Llama-3.1-70B-Instruct


  meta-llama/Meta-Llama-3.1-405B
  meta-llama/Meta-Llama-3.1-405B-FP8
  meta-llama/Meta-Llama-3.1-405B-Instruct
  meta-llama/Meta-Llama-3.1-405B-Instruct-FP8


Testing:
  A. Launch OnDemandTextInput with Explorer
  B. Write custom command (see below)



for llama3.1 in-filling:
```json
{
  "ACTION" : "PIPELINE_COMMAND",
  "PAYLOAD" : {
    "NAME": "llm_request",
    "PIPELINE_COMMAND" : {
      "STRUCT_DATA" : {
        "request" : "What is the square root of 4?",
        "history" : [
          {
            "request"   : "hello",
            "response"  : "Hello, how can I help you today?"
          }
        ],
        "system_info" : "You are a funny university teacher. Your task is to help students with their learning journey."
      }
    }
  }
}
```
"""

from naeural_core.serving.base.base_llm_serving import BaseLlmServing as BaseServingProcess

__VER__ = '0.1.0.0'

_CONFIG = {
  **BaseServingProcess.CONFIG,

  "MODEL_NAME": "meta-llama/Meta-Llama-3.1-8B-Instruct",

  "PICKED_INPUT": "STRUCT_DATA",
  "RUNS_ON_EMPTY_INPUT": False,

  'VALIDATION_RULES': {
    **BaseServingProcess.CONFIG['VALIDATION_RULES'],
  },

}


class LlamaV31(BaseServingProcess):

  def _setup_llm(self):
    # just override this method as the base class has a virtual method that raises an exception
    return



import os
import math
import numpy as np
import matplotlib.pyplot as plt
from tqdm import tqdm
from PIL import Image
from enum import Enum
from abc import abstractmethod
from naeural_client import BaseDecentrAIObject


__version__ = '1.0.1.1'
__VER__ = __version__


class VisEngineEnum(Enum):
  CV2 = 0
  PIL = 1

class RGBColorEnum(Enum):
  DEFAULT_BLACK     = (0, 0, 0)
  DEFAULT_WHITE     = (255, 255, 255)
  DEFAULT_RED       = (255, 0, 0)
  DEFAULT_GREEN     = (0, 255, 0)
  DEFAULT_BLUE      = (0, 0, 255)
  
  DARK_BLUE    = (11, 45, 109)
  LIGHT_BLUE   = (129, 255, 254)
  ORANGE       = (250, 149, 19)
  YELLOW       = (250, 222, 98)
  RED          = (248, 70, 100)
  GREEN        = (50, 200, 50)

IMG_EXT = ['.png', '.bmp', '.jpg', '.jpeg']

class BaseVisualization(BaseDecentrAIObject):
  def __init__(self, **kwargs):
    super().__init__(**kwargs)
    return
  
  @abstractmethod
  def display(self, img, window_name):
    raise NotImplementedError()
  
  @abstractmethod
  def close_display(self):
    raise NotImplementedError()
  
  @abstractmethod  
  def draw_rect(self, img, top, left, bottom, right, color, thickness):
    raise NotImplementedError()
  
  @abstractmethod
  def draw_box(self, img, top, left, bottom, right, display_str, color, thickness):
    raise NotImplementedError()
  
  def draw_boxes(self, img, lst_boxes, lst_display_str, lst_color, thickness):
    raise NotImplementedError()
    
  @abstractmethod
  def draw_text(self, img, text, left, top, color, thickness, font_scale):
    raise NotImplementedError() 
  
  @abstractmethod
  def read_image(self, path):
    raise NotImplementedError()
    
  @abstractmethod
  def save_image(self, path, img):
    raise NotImplementedError()
    
  def resize_image(self, height, width, interpolation):
    raise NotImplementedError()
  
  def is_image(self, path):
    return any(path.endswith(x) for x in IMG_EXT)
  
  def read_images(self, fn):
    if isinstance(fn, str) and os.path.isdir(fn):
      lst_files = [x for x in os.listdir(fn) if self.is_image(x)]
      lst_paths = [os.path.join(fn, x) for x in lst_files]
    elif isinstance(fn, str):
      if os.path.exists(fn) and self.is_image(fn):
        lst_paths = [fn]
      else:
        raise ValueError('File named incorrect: {}'.format(fn))
    elif isinstance(fn, list):
      lst_paths = [x for x in fn if self.is_image(x)]
    else:
      raise ValueError('Value not understood: {}'.format(fn))
    lst_bgr = [self.read_image(x) for x in tqdm(lst_paths)]
    lst_rgb = [self.reverse_channels(x) for x in lst_bgr]
    return lst_paths, lst_bgr, lst_rgb

  def save_images(self, lst_imgs, reverse_channels, path=None):
    self.log.p('Saving images')
    if path is None:
      path = self.log.now_str()
    os.makedirs(path, exist_ok=True)
    for idx, img in tqdm(enumerate(lst_imgs)):
      if reverse_channels:
        img = self.reverse_channels(img)
      self.save_image(os.path.join(path, '{}.png'.format(idx)), img)
    self.log.p('Done saving images', show_time=True)
    return

  def show(self, img, title=None, xlabel=None, ylabel=None):
    plt.imshow(img)
    if title:
      plt.title(title)
    if xlabel:
      plt.xlabel(xlabel)
    if ylabel:
      plt.ylabel(ylabel)
    plt.show()
    return

  def side_by_side(self, lst_imgs, lst_titles=None, lst_xlabel=None, lst_ylabel=None, title=None):
    for img in lst_imgs:
      if img.max() <= 1: assert img.dtype in ['float32', 'float64']
      if img.max() > 1 : assert img.dtype == 'uint8'
    
    if lst_titles:
      assert len(lst_imgs) == len(lst_titles)
    if lst_xlabel:
      assert len(lst_imgs) == len(lst_xlabel)
    if lst_ylabel:
      assert len(lst_imgs) == len(lst_ylabel)
    
    nr = len(lst_imgs)
    for idx, img in enumerate(lst_imgs):
      plt.subplot(1, nr, idx+1)
      if lst_titles: plt.title(lst_titles[idx])
      if lst_xlabel: plt.xlabel(lst_xlabel[idx])
      if lst_ylabel: plt.ylabel(lst_ylabel[idx])
      plt.imshow(img)
    if title:
      plt.title(title)
    plt.show()
    return
  
  def display_images(self, imgs, n_cols, 
                     titles=None, nr_max=20, 
                     axis_off=False, title=None,
                     figsize=(13,8)):
    if titles:
      assert len(imgs) == len(titles)
    
    if nr_max is not None:
      imgs = imgs[:nr_max]    
    n_rows = int(math.ceil(len(imgs) / n_cols))      
    crt_img = 0
    fig, axs = plt.subplots(n_rows, n_cols, figsize=figsize)
    for row in range(n_rows):
      for col in range(n_cols):
        idx = row * n_cols + col
        if idx >= len(imgs):
          img = np.zeros(shape=(10, 10, 3), dtype=np.uint8)
        else:
          img = imgs[idx]
        if len(axs.shape) == 1:
          axs[col].imshow(img)
        else:
          axs[row, col].imshow(img)
        if titles:
          lbl = ''
          if idx < len(titles):
            lbl = titles[idx]
          if len(axs.shape) == 1:
            axs[col].set_title(lbl)
          else:
            axs[row, col].set_title(lbl)
        if axis_off:
          axs[row, col].axis('off')
        crt_img+= 1
        if crt_img >= (n_rows * n_cols):
          break
      #endfor
      if crt_img >= (n_rows * n_cols):
        break
    if title:
      fig.suptitle(title)
    plt.show()    
    return
  
  def display_in_batches(self, imgs, n_cols, batch_size=16, titles=None, title=None):
    if titles:
      assert len(imgs) == len(titles)
    
    nr_batches = int(math.ceil(len(imgs) / batch_size))
    for nr in tqdm(range(nr_batches)):
      start = batch_size * nr
      stop = start + batch_size if nr < nr_batches - 1 else len(imgs)
      x_batch = imgs[start:stop]
      if titles:
        y_batch = titles[start:stop]
      else:
        y_batch = None
      plot_title = '{} batch {}'.format(title, nr + 1) if title else 'Batch {}'.format(nr + 1)
      self.display_images(x_batch, n_cols=n_cols, titles=y_batch, title=plot_title)
    return
  
  def reverse_channels(self, img):
    img = img[:, :, ::-1]
    return img
      
  def transform_image(self, img, normalize, reverse_channels):
    if normalize:
      img/= 255
    if reverse_channels:
      img = self.reverse_channels(img)
    return img  
  
  def reverse_color(self, color):
    return color[::-1]
  
  def is_image(self, img_path):
    return any([img_path.endswith(x) for x in IMG_EXT])
  
  def plot_distribution(self, classes):
    assert isinstance(classes, (list, np.ndarray))
    np_unique, np_counts = np.unique(classes, return_counts=True)
    plt.bar(np.arange(len(np_unique)), np_counts)
    plt.xticks(np_unique)
    plt.title('Class distribution')
    plt.xlabel('Class')
    plt.ylabel('Nr occurance')
    plt.show()
    return
  
  
  
class VisualizationCV2(BaseVisualization):
  def __init__(self, **kwargs):
    super().__init__(**kwargs)
    import cv2
    self.__cv2 = cv2
    self.font = kwargs.get('font', self.__cv2.FONT_HERSHEY_COMPLEX)
    return
  
  def close_display(self):
    self.__cv2.waitKey(1)
    for i in range(5):
      self.__cv2.waitKey(1)
      self.__cv2.destroyAllWindows()
    return
  
  def show(self, img, window_name='Debug'):
    if isinstance(img, str):
      img = self.read_image(img)
    self.__cv2.imshow(window_name, img)
    return
  
  def display(self, img, window_name='Debug'):
    self.show(img, window_name)
    self.__cv2.waitKey(0)
    self.close_display()
    return
  
  def read_image(self, path):
    assert os.path.exists(path)
    return self.__cv2.imread(path)
  
  def save_image(self, path, img):
    self.__cv2.imwrite(path, img)
    return
  
  def resize_image(self, img, height, width, interpolation=None):
    if interpolation is None:
      interpolation = self.__cv2.INTER_LINEAR
    dim = (width, height)
    img_res = self.__cv2.resize(img, dim, interpolation=interpolation)
    return img_res
  
  def draw_text(self, img, text, left, top, color=RGBColorEnum.DEFAULT_BLACK.value, thickness=1, font_scale=1):
    self.__cv2.putText(img=img, text=text, org=(left, top), color=color, thickness=thickness, 
                fontScale=font_scale, fontFace=self.font)
    return
  
  def draw_rect(self, img, top, left, bottom, right, color=RGBColorEnum.DEFAULT_GREEN.value, thickness=2):
    self.__cv2.rectangle(img=img, pt1=(left, top), pt2=(right, bottom), color=color, thickness=thickness)
    return
  
  def draw_box(self, img, top, left, bottom, right, display_str='', color=RGBColorEnum.DEFAULT_GREEN.value, thickness=2):
    #bounding box
    self.draw_rect(img=img, top=top, left=left, bottom=bottom, right=right, color=color, thickness=thickness)
    
    if display_str != '':
      font_scale = 0.6
      font_thickness = 1
      (tw, th) = self.__cv2.getTextSize(text=display_str, fontFace=self.font, fontScale=font_scale, thickness=font_thickness)[0]
      top_desc = top - th
      bottom_desc = top
      left_desc = left
      right_desc = left_desc + tw
      self.draw_rect(img=img, top=top_desc, left=left_desc, bottom=bottom_desc, right=right_desc, color=color, thickness=self.__cv2.FILLED)
      self.draw_text(img=img, text=display_str, left=left_desc, top=bottom_desc, color=RGBColorEnum.DEFAULT_BLACK.value,
                     thickness=font_thickness, font_scale=font_scale)
    #endif
    return
  
  def draw_boxes(self, img, lst_boxes, lst_display_str, lst_color=None, thickness=2):
    if lst_color is None:
      lst_color = [tuple([int(x) for x in np.random.randint(low=0, high=255, size=3)]) for _ in range(len(lst_boxes))]
    
    assert len(lst_boxes) == len(lst_display_str) == len(lst_color)        
    for box, display_str, color in zip(lst_boxes, lst_display_str, lst_color):
      top, left, bottom, right = box
      self.draw_box(img=img, top=top, left=left, bottom=bottom, right=right,
                    display_str=display_str, color=color, thickness=thickness)
    return img
  
  def draw_inferences(self, img, lst_preds, lst_color=None):
    lst_boxes = [x['TLBR_POS'] for x in lst_preds]
    lst_display_str = ['{}: {:.2f}%'.format(x['TYPE'], x['PROB_PRC']) for x in lst_preds]
    self.draw_boxes(
      img=img,
      lst_boxes=lst_boxes,
      lst_display_str=lst_display_str,
      lst_color=lst_color
      )
    return img
  
class VisualizationPIL(BaseVisualization):
  def __init__(self, **kwargs):
    super().__init__(**kwargs)
    return
  
  def close_display(self):
    self.log.p('PIL opens images in separate processes. Please parse process and close the opend image')
    return
  
  def show(self, img, window_name='Debug'):
    if isinstance(img, str):
      img = self.read_image(img)
    if isinstance(img, np.ndarray):
      img = Image.fromarray(img)
    img.show()
    return
  
  def display(self, img, window_name='Debug'):
    self.show(img, window_name)
    return
  
  def read_image(self, path, return_numpy=True):
    assert os.path.exists(path)
    img = Image.open(path)
    if return_numpy:
      img = np.asarray(img)
    return img
  
  def save_image(self, path, img):
    if isinstance(img, np.ndarray):
      img = Image.fromarray(img)
    elif isinstance(img, Image):
      pass
    else:
      raise ValueError('Please provide either np image or PIL Image')
    img.save(path)
    return
  
  def resize_image(self, img, height, width, interpolation=Image.BILINEAR):
    size = (width, height)
    if isinstance(img, np.ndarray):
      img = Image.fromarray(img)
    img_res = img.resize(size=size, resample=interpolation)
    img_res = np.asarray(img_res)
    return img_res
  
  
class VisEngine(BaseDecentrAIObject):
  def __init__(self, vis_engine=VisEngineEnum.CV2, **kwargs):
    super().__init__(**kwargs)    
    self._init_eng(vis_engine, **kwargs)
    return
  
  def _init_eng(self, vis_engine, **kwargs):
    if vis_engine == VisEngineEnum.CV2:
      self.vis_eng = VisualizationCV2(**kwargs)
    else:
      self.vis_eng = VisualizationPIL(**kwargs)
    return
  
  def get_eng(self):
    return self.vis_eng

if __name__ == '__main__':
  from naeural_core import Logger
  log = Logger(lib_name='TST', config_file='config.txt')
  
  vis = VisEngine(log=log, vis_engine=VisEngineEnum.PIL).get_eng()
  img = vis.read_image('C:/Users/ETA/Dropbox/DATA/_vapor_data/_cervical_lessions/_data/train/0/20150722014/20150722163342.jpg')
  vis.show(img)
  vis.save_image('a.png', img)
  
  
  
  
  
  
  


"""
@misc{touvron2023llama,
      title={Llama 2: Open Foundation and Fine-Tuned Chat Models}, 
      author={Hugo Touvron and Louis Martin and Kevin Stone and Peter Albert and Amjad Almahairi and Yasmine Babaei and Nikolay Bashlykov and Soumya Batra and Prajjwal Bhargava and Shruti Bhosale and Dan Bikel and Lukas Blecher and Cristian Canton Ferrer and Moya Chen and Guillem Cucurull and David Esiobu and Jude Fernandes and Jeremy Fu and Wenyin Fu and Brian Fuller and Cynthia Gao and Vedanuj Goswami and Naman Goyal and Anthony Hartshorn and Saghar Hosseini and Rui Hou and Hakan Inan and Marcin Kardas and Viktor Kerkez and Madian Khabsa and Isabel Kloumann and Artem Korenev and Punit Singh Koura and Marie-Anne Lachaux and Thibaut Lavril and Jenya Lee and Diana Liskovich and Yinghai Lu and Yuning Mao and Xavier Martinet and Todor Mihaylov and Pushkar Mishra and Igor Molybog and Yixin Nie and Andrew Poulton and Jeremy Reizenstein and Rashi Rungta and Kalyan Saladi and Alan Schelten and Ruan Silva and Eric Michael Smith and Ranjan Subramanian and Xiaoqing Ellen Tan and Binh Tang and Ross Taylor and Adina Williams and Jian Xiang Kuan and Puxin Xu and Zheng Yan and Iliyan Zarov and Yuchen Zhang and Angela Fan and Melanie Kambadur and Sharan Narang and Aurelien Rodriguez and Robert Stojnic and Sergey Edunov and Thomas Scialom},
      year={2023},
      eprint={2307.09288},
      archivePrefix={arXiv},
      primaryClass={cs.CL}
}


@misc{rozière2023code,
      title={Code Llama: Open Foundation Models for Code}, 
      author={Baptiste Rozière and Jonas Gehring and Fabian Gloeckle and Sten Sootla and Itai Gat and Xiaoqing Ellen Tan and Yossi Adi and Jingyu Liu and Tal Remez and Jérémy Rapin and Artyom Kozhevnikov and Ivan Evtimov and Joanna Bitton and Manish Bhatt and Cristian Canton Ferrer and Aaron Grattafiori and Wenhan Xiong and Alexandre Défossez and Jade Copet and Faisal Azhar and Hugo Touvron and Louis Martin and Nicolas Usunier and Thomas Scialom and Gabriel Synnaeve},
      year={2023},
      eprint={2308.12950},
      archivePrefix={arXiv},
      primaryClass={cs.CL}
}


models:
  codellama/CodeLlama-7b-hf
  codellama/CodeLlama-7b-Python-hf
  codellama/CodeLlama-7b-Instruct-hf
  
  codellama/CodeLlama-13b-hf
  codellama/CodeLlama-13b-Python-hf
  codellama/CodeLlama-13b-Instruct-hf
  
  
  codellama/CodeLlama-34b-hf
  codellama/CodeLlama-34b-Python-hf
  codellama/CodeLlama-34b-Instruct-hf    
  

Llama 2 (classic): "meta-llama/Llama-2-7b-chat-hf"




Testing:
  A. Launch OnDemandTextInput with Explorer
  B. Write custom command (see below)
  
  

for code llama in-filling:
```json
{ 
  "ACTION" : "PIPELINE_COMMAND",  
  "PAYLOAD" : {
    "NAME": "code-on-demand",
    "PIPELINE_COMMAND" : {
      "STRUCT_DATA" : {
        "request" : "def hello_world(<FILL>\n\nhello_world()",
        "history" : [
          {
            "request"   : "print('hello",
            "response"  : " world')"            
          }
        ],        
        "system_info" : "You are a funny python programmer assistant. your task is to complete the code you are given. return only the completion, not the whole program."
      }
    }
  }
}
```
TODO:
0. Prepare python SDK script for testing.
1. Run with 16 bit, observe allocation and inference time.
2. Run with 8 bit, observe allocation and inference time.
3. Run with 4 bit, observe allocation and inference time.
"""



from naeural_core.serving.base.base_llm_serving import BaseLlmServing as BaseServingProcess


__VER__ = '0.1.0.0'

_CONFIG = {
  **BaseServingProcess.CONFIG,
  
  # this was used before CodeLlamaTokenizer was available in transformers, make sure you use 4.34 or above
  # "MODEL_NAME" : "meta-llama/Llama-2-7b-chat-hf", 
  # end generic tokenizer use
  
  "MODEL_NAME" : "codellama/CodeLlama-7b-Instruct-hf",

  "PICKED_INPUT" : "STRUCT_DATA",
  
  "RUNS_ON_EMPTY_INPUT" : False,

  'VALIDATION_RULES': {
    **BaseServingProcess.CONFIG['VALIDATION_RULES'],
  },

}

class CodeLlamaV2(BaseServingProcess):
  def __init__(self, **kwargs):
    self._counter = 0
    super(CodeLlamaV2, self).__init__(**kwargs)
    return
  
  def _setup_llm(self):
    # just override this method as the base class has a virtual method that raises an exception
    return  



"""
@misc{touvron2023llamaopenefficientfoundation,
      title={LLaMA: Open and Efficient Foundation Language Models},
      author={Hugo Touvron and Thibaut Lavril and Gautier Izacard and Xavier Martinet and Marie-Anne Lachaux and Timothée Lacroix and Baptiste Rozière and Naman Goyal and Eric Hambro and Faisal Azhar and Aurelien Rodriguez and Armand Joulin and Edouard Grave and Guillaume Lample},
      year={2023},
      eprint={2302.13971},
      archivePrefix={arXiv},
      primaryClass={cs.CL},
      url={https://arxiv.org/abs/2302.13971},
}

@misc{masala2024vorbecstiromanecsterecipetrain,
      title={"Vorbe\c{s}ti Rom\^ane\c{s}te?" A Recipe to Train Powerful Romanian LLMs with English Instructions},
      author={Mihai Masala and Denis C. Ilie-Ablachim and Alexandru Dima and Dragos Corlatescu and Miruna Zavelca and Ovio Olaru and Simina Terian and Andrei Terian and Marius Leordeanu and Horia Velicu and Marius Popescu and Mihai Dascalu and Traian Rebedea},
      year={2024},
      eprint={2406.18266},
      archivePrefix={arXiv},
      primaryClass={cs.CL},
      url={https://arxiv.org/abs/2406.18266},
}

models:
  pansophic/pansophic-1-preview-LLaMA3.1-8b
  pansophic/pansophic-1-preview-LLaMA3.1-8b-GPTQ
  pansophic/pansophic-1-preview-LLaMA3.1-8b-GGUF

Testing:
  A. Launch OnDemandTextInput with Explorer
  B. Write custom command (see below)



for llama3.1 in-filling:
```json
{
  "ACTION" : "PIPELINE_COMMAND",
  "PAYLOAD" : {
    "NAME": "llm_request",
    "PIPELINE_COMMAND" : {
      "STRUCT_DATA" : {
        "request" : "What is the square root of 4?",
        "history" : [
          {
            "request"   : "hello",
            "response"  : "Hello, how can I help you today?"
          }
        ],
        "system_info" : "You are a funny university teacher. Your task is to help students with their learning journey."
      }
    }
  }
}
```
"""

from naeural_core.serving.base.base_llm_serving import BaseLlmServing as BaseServingProcess

__VER__ = '0.1.0.0'

_CONFIG = {
  **BaseServingProcess.CONFIG,

  # "MODEL_NAME": "pansophic/pansophic-1-preview-LLaMA3.1-8b",
  "MODEL_NAME": "OpenLLM-Ro/RoLlama3.1-8b-Instruct-DPO",

  "PICKED_INPUT": "STRUCT_DATA",
  "RUNS_ON_EMPTY_INPUT": False,

  'VALIDATION_RULES': {
    **BaseServingProcess.CONFIG['VALIDATION_RULES'],
  },

}


class RoLlamaV31(BaseServingProcess):

  def add_context_to_request(self, request, context):
    return f"{request} - Te rog sa raspunzi luand in considerare si urmatoarele: {context}"

  def _setup_llm(self):
    # just override this method as the base class has a virtual method that raises an exception
    return



import os
import shutil
import subprocess
import tempfile
import asyncio

from jinja2 import Environment, FileSystemLoader

from naeural_core.business.base import BasePluginExecutor
from naeural_core.business.mixins_libs.ngrok_mixin import _NgrokMixinPlugin

__VER__ = '0.0.0.0'

_CONFIG = {
  **BasePluginExecutor.CONFIG,
  'ALLOW_EMPTY_INPUTS': True,
  'RUN_WITHOUT_IMAGE': True,
  'PROCESS_DELAY': 5,
  
  'GIT_REQUEST_DELAY': 60 * 10,  # 10 minutes

  'NGROK_USE_API': True,
  'NGROK_ENABLED': False,
  'NGROK_DOMAIN': None,
  'NGROK_EDGE_LABEL': None,
  'NGROK_URL_PING_INTERVAL': 30,

  'ASSETS': None,

  'SETUP_COMMANDS': [],
  'START_COMMANDS': [],
  'ENV_VARS': {},
  'AUTO_START': True,
  'FORCED_RELOAD_INTERVAL': None,

  'PORT': None,

  'VALIDATION_RULES': {
    **BasePluginExecutor.CONFIG['VALIDATION_RULES']
  },
}


class BaseWebAppPlugin(_NgrokMixinPlugin, BasePluginExecutor):
  """
  A base plugin which will handle the lifecycle of a web application.
  Through this plugin, you can expose your business logic as a web application,
  using some implementation of a web server.

  You can also deploy your web application to the internet using ngrok.
  To do this, set the `NGROK_ENABLED` flag to True in config and set the necessary
  environment variables.

  TODO: add ngrok necessary data in the config (after securing the configs)
  """

  CONFIG = _CONFIG
  
  def _on_init(self):

    self.__git_commit_hash = None
    self.__git_request_time = 0
    self.ngrok_initiated = False
    self.ngrok_started = False
    self.ngrok_listener = None
    self.__last_ngrok_url_ping_ts = 0

    # TODO: move this to process
    self.__allocate_port()

    self.prepared_env = None
    self.base_env = None

    self.script_temp_dir = tempfile.mkdtemp()

    self.assets_initialized = False
    self.failed = False
    self.__commands_ready = False

    self.can_run_start_commands = self.cfg_auto_start
    self.setup_commands_started = []
    self.setup_commands_finished = []
    self.setup_commands_processes = []
    self.setup_commands_start_time = []

    self.start_commands_started = []
    self.start_commands_finished = []
    self.start_commands_processes = []
    self.start_commands_start_time = []
    self.webapp_reload_last_timestamp = 0

    super(BaseWebAppPlugin, self)._on_init()
    return
  
  
  def __can_request_git(self):
    if self.time() - self.__git_request_time > self.cfg_git_request_delay:
      self.__git_request_time = self.time()
      return True
    return False


  # Port allocation
  def __check_port_valid(self):
    # Check the config as we're going to use it to start processes.
    if not isinstance(self.cfg_port, int):
      raise ValueError("Port not an int")
    if self.cfg_port < 0 or self.cfg_port > 65535:
      raise ValueError("Invalid port value {}".format(self.cfg_port))
    return

  def __allocate_port(self):
    with self.managed_lock_resource('USED_PORTS'):
      if 'USED_PORTS' not in self.plugins_shmem:
        self.plugins_shmem['USED_PORTS'] = {}
      dct_shmem_ports = self.plugins_shmem['USED_PORTS']
      used_ports = dct_shmem_ports.values()

      if self.cfg_port is not None:
        self.__check_port_valid()

        if self.cfg_port in used_ports:
          raise Exception("Port {} is already in use.".format(self.cfg_port))
        else:
          dct_shmem_ports[self.str_unique_identification] = self.cfg_port
      else:
        port = self.np.random.randint(49152, 65535)
        while port in used_ports:
          port = self.np.random.randint(49152, 65535)
        # endwhile
        dct_shmem_ports[self.str_unique_identification] = port
      # endif port
    # endwith lock
    return

  def __deallocate_port(self):
    port = self.port

    with self.managed_lock_resource('USED_PORTS'):
      if 'USED_PORTS' in self.plugins_shmem:
        self.plugins_shmem['USED_PORTS'].pop(self.str_unique_identification, None)
    # endwith lock

    self.P(f"Released port {port}")
    return

  def __prepare_env(self, assets_path):
    # pop all `EE_` keys
    prepared_env = dict(self.os_environ)
    to_pop_keys = []
    for key in prepared_env:
      if key.startswith('EE_'):
        to_pop_keys.append(key)
    # endfor all keys

    for key in to_pop_keys:
      prepared_env.pop(key)

    # add mandatory keys
    prepared_env["PWD"] = self.script_temp_dir
    prepared_env["PORT"] = str(self.port)

    self.base_env = prepared_env.copy()

    # add optional keys, found in `.env` file from assets folder if provided.
    if assets_path is not None:
      env_file_path = self.os_path.join(assets_path, '.env')
      if self.os_path.exists(env_file_path):
        with open(env_file_path, 'r') as f:
          for line in f:
            if line.startswith('#') or len(line.strip()) == 0:
              continue
            key, value = line.strip().split('=', 1)
            prepared_env[key] = value
        # endwith

        # TODO: should we remove the .env file?
      # endif env file
    # endif assets path provided

    # add optional keys, found in config
    if self.cfg_env_vars is not None and isinstance(self.cfg_env_vars, dict):
      processed_env = {k: str(v) for k, v in self.cfg_env_vars.items()}
      prepared_env.update(processed_env)

    self.prepared_env = prepared_env

    return prepared_env

  # process handling methods
  def __run_command(self, command, env=None, read_stdout=True, read_stderr=True):
    if isinstance(command, list):
      command_args = command
    elif isinstance(command, str):
      command_args = command.split(' ')
    else:
      raise ValueError("Command must be a string or a list of strings")

    process = subprocess.Popen(
      command_args,
      env=env,
      cwd=self.script_temp_dir,
      stdout=subprocess.PIPE if read_stdout else None,
      stderr=subprocess.PIPE if read_stderr else None,
    )
    logs_reader = self.LogReader(process.stdout) if read_stdout else None
    err_logs_reader = self.LogReader(process.stderr) if read_stderr else None
    return process, logs_reader, err_logs_reader

  def __wait_for_command(self, process, timeout):
    process_finished = False
    failed = False
    try:
      process.wait(timeout)
      failed = process.returncode != 0
      process_finished = True
    except subprocess.TimeoutExpired:
      pass

    return process_finished, failed

  def __maybe_kill_process(self, process, key, max_tries=5):
    if process is None:
      return
    tries = 0
    success = False
    while tries < max_tries and not success:
      tries += 1
      try:
        self.P(f"Forcefully killing process {key}(try {tries}/{max_tries})")
        process.kill()
        success, failed = self.__wait_for_command(
          process=process,
          timeout=3,
        )
        self.__maybe_print_key_logs(key)
        if success:
          self.P(f"Killed process {key} from {tries} tries.")
      except Exception as exc:
        self.P(f'Could not kill process {key} (try {tries}/{max_tries}). Reason: {exc}')
    # endwhile
    return
  

  # logs handling methods
  def __maybe_print_all_logs(self, indent=35):
    for key, logs_reader in self.dct_logs_reader.items():
      if logs_reader is not None:
        logs = logs_reader.get_next_characters()
        if len(logs) > 0:
          indented_logs = self.indent_strings(logs, indent=indent)
          self.P(f"Showing stdout logs [{key}]:\n{indented_logs}")
          self.logs.append(f"[{key}]: {logs}")

    for key, err_logs_reader in self.dct_err_logs_reader.items():
      if err_logs_reader is not None:
        err_logs = err_logs_reader.get_next_characters()
        if len(err_logs) > 0:
          indented_err_logs = self.indent_strings(err_logs, indent=indent)
          self.P(f"Showing error logs [{key}]:\n{indented_err_logs}")
          self.err_logs.append(f"[{key}]: {err_logs}")
    return


  def __maybe_print_key_logs(self, key):
    logs_reader = self.dct_logs_reader.get(key)
    if logs_reader is not None:
      logs = logs_reader.get_next_characters()
      if len(logs) > 0:
        self.P(f"[{key}]: {logs}")
        self.logs.append(f"[{key}]: {logs}")

    err_logs_reader = self.dct_err_logs_reader.get(key)
    if err_logs_reader is not None:
      err_logs = err_logs_reader.get_next_characters()
      if len(err_logs) > 0:
        self.P(f"[{key}]: {err_logs}")
        self.err_logs.append(f"[{key}]: {err_logs}")
    return


  def __get_delta_logs(self):
    logs = list(self.logs)
    logs = "".join(logs)
    self.logs.clear()

    err_logs = list(self.err_logs)
    err_logs = "".join(err_logs)
    self.err_logs.clear()

    return logs, err_logs

  def __maybe_read_and_stop_key_log_readers(self, key):
    logs_reader = self.dct_logs_reader.get(key)
    if logs_reader is not None:
      logs_reader.stop()

    err_logs_reader = self.dct_err_logs_reader.get(key)
    if err_logs_reader is not None:
      err_logs_reader.stop()

    self.__maybe_print_key_logs(key)

    self.dct_logs_reader.pop(key, None)
    self.dct_err_logs_reader.pop(key, None)
    return

  def __maybe_read_and_stop_all_log_readers(self):
    log_keys = set(list(self.dct_logs_reader.keys()) + list(self.dct_err_logs_reader.keys()))
    for key in log_keys:
      self.__maybe_read_and_stop_key_log_readers(key)
    return

  # setup commands methods
  def __maybe_run_all_setup_commands(self):
    if self.failed:
      return

    has_finished_setup_commands = self.__has_finished_setup_commands()
    if has_finished_setup_commands:
      return
    
    self.P("Running setup commands (`has_finished_setup_commands: {}`)".format(has_finished_setup_commands))

    for idx in range(len(self.get_setup_commands())):
      self.__maybe_run_nth_setup_command(idx)
    return


  def __maybe_close_setup_commands(self):
    if not any(self.setup_commands_started):
      self.P("No setup commands were started. Skipping teardown.")
      return

    if all(self.setup_commands_finished):
      self.P("All setup commands have finished. Skipping teardown.")
      return

    for idx, process in enumerate(self.setup_commands_processes):
      if self.setup_commands_started[idx] and not self.setup_commands_finished[idx]:
        self.P(f"Setup command nr {idx} has not finished. Killing it.")
        self.__maybe_kill_process(process, f"setup_{idx}")
    # endfor all setup commands
    return


  def __maybe_run_nth_setup_command(self, idx, timeout=None):
    if self.failed:
      return

    if idx > 0 and not self.setup_commands_finished[idx - 1]:
      # Previous setup command has not finished yet. Skip this one.
      return

    if not self.setup_commands_started[idx]:
      self.P(f"Running setup command nr {idx}: {self.get_setup_commands()[idx]}")
      proc, logs_reader, err_logs_reader = self.__run_command(self.get_setup_commands()[idx], self.base_env)
      self.setup_commands_processes[idx] = proc
      self.dct_logs_reader[f"setup_{idx}"] = logs_reader
      self.dct_err_logs_reader[f"setup_{idx}"] = err_logs_reader

      self.setup_commands_started[idx] = True
      self.setup_commands_start_time[idx] = self.time()
    # endif setup command started

    if not self.setup_commands_finished[idx]:
      finished, failed = self.__wait_for_command(
        process=self.setup_commands_processes[idx],
        timeout=0.1,
      )

      self.setup_commands_finished[idx] = finished

      if finished and not failed:
        self.__maybe_read_and_stop_key_log_readers(f"setup_{idx}")
        self.add_payload_by_fields(
          command_type="setup",
          command_idx=idx,
          command_str=self.get_setup_commands()[idx],
          command_status="success"
        )
        self.P(f"Setup command nr {idx} finished successfully")
      elif finished and failed:
        self.__maybe_read_and_stop_key_log_readers(f"setup_{idx}")
        self.add_payload_by_fields(
          command_type="setup",
          command_idx=idx,
          command_str=self.get_setup_commands()[idx],
          command_status="failed"
        )
        self.P(f"ERROR: Setup command nr {idx} finished with exit code {self.setup_commands_processes[idx].returncode}")
        self.failed = True
      elif not finished and timeout is not None and timeout > 0:
        if self.time() - self.setup_commands_start_time[idx] > timeout:
          self.setup_commands_processes[idx].kill()
          self.__maybe_read_and_stop_key_log_readers(f"setup_{idx}")
          self.P(f"ERROR: Setup command nr {idx} timed out")
          self.add_payload_by_fields(
            command_type="setup",
            command_idx=idx,
            command_str=self.get_setup_commands()[idx],
            command_status="timeout"
          )
          self.failed = True
    # endif setup command finished
    return


  def __has_finished_setup_commands(self):
    return all(self.setup_commands_finished)


  # start commands methods
  def __maybe_run_all_start_commands(self):
    if self.failed:
      return

    if self.__has_finished_start_commands():
      return

    if not self.__has_finished_setup_commands():
      return

    if not self.can_run_start_commands:
      return
    
    self.P("Running START commands...")

    for idx in range(len(self.get_start_commands())):
      self.__maybe_run_nth_start_command(idx)
    return

  def __maybe_close_start_commands(self):
    if not any(self.start_commands_started):
      self.P("Server was never started. Skipping teardown.")
      return

    for idx, process in enumerate(self.start_commands_processes):
      self.__maybe_kill_process(process, f"start_{idx}")
    # endfor all start commands

  def __maybe_run_nth_start_command(self, idx, timeout=5):
    if self.failed:
      return

    if idx > 0 and not self.start_commands_finished[idx - 1]:
      # Previous start command has not finished yet. Skip this one.
      return

    if not self.start_commands_started[idx]:
      self.P(f"Running start command nr {idx}: {self.get_start_commands()[idx]}")
      proc, logs_reader, err_logs_reader = self.__run_command(self.get_start_commands()[idx], self.prepared_env)
      self.start_commands_processes[idx] = proc
      self.dct_logs_reader[f"start_{idx}"] = logs_reader
      self.dct_err_logs_reader[f"start_{idx}"] = err_logs_reader

      self.start_commands_started[idx] = True
      self.start_commands_start_time[idx] = self.time()
    # endif start command started

    if not self.start_commands_finished[idx]:
      finished, _ = self.__wait_for_command(
        process=self.start_commands_processes[idx],
        timeout=0.1,
      )

      self.start_commands_finished[idx] = finished

      if finished:
        self.__maybe_read_and_stop_key_log_readers(f"start_{idx}")
        self.add_payload_by_fields(
          command_type="start",
          command_idx=idx,
          command_str=self.get_start_commands()[idx],
          command_status="failed"
        )
        self.P(f"Start command nr {idx} finished unexpectedly. Please check the logs.")
        self.failed = True
      elif self.time() - self.start_commands_start_time[idx] > timeout:
        self.start_commands_finished[idx] = True
        self.add_payload_by_fields(
          command_type="start",
          command_idx=idx,
          command_str=self.get_start_commands()[idx],
          command_status="success"
        )
        self.P(f"Start command nr {idx} is running")
    # endif setup command finished
    return

  def __has_finished_start_commands(self):
    return all(self.start_commands_finished)
  
  
  def __reload_server(self):
    self.P("Initiating server reload...")
    self.__maybe_close_setup_commands()
    self.__maybe_close_start_commands()
    self.__maybe_read_and_stop_all_log_readers()

    self.assets_initialized = False
    self.failed = False

    self.setup_commands_started = [False] * len(self.get_setup_commands())
    self.setup_commands_finished = [False] * len(self.get_setup_commands())
    self.setup_commands_processes = [None] * len(self.get_setup_commands())
    self.setup_commands_start_time = [None] * len(self.get_setup_commands())

    self.start_commands_started = [False] * len(self.get_start_commands())
    self.start_commands_finished = [False] * len(self.get_start_commands())
    self.start_commands_processes = [None] * len(self.get_start_commands())
    self.start_commands_start_time = [None] * len(self.get_start_commands())

    self.P('Attempting to init assets due to reload...')
    self.__maybe_init_assets()
    return  
  
  
  
  def __check_new_repo_version(self):
    result = False # return false by default including if no git url is provided
    # first check not to run this operation too many times
    if isinstance(self.cfg_assets, dict) and self.cfg_assets.get('operation') == 'clone':
      url = self.cfg_assets.get('url')
      username = self.cfg_assets.get('username')
      token = self.cfg_assets.get('token')
      if self.__git_commit_hash is not None:
        # check if we can request git based on a configured delay
        can_request_git = self.__can_request_git()
        if can_request_git:
          commit_hash = self.git_get_last_commit_hash(
            repo_url=url,
            user=username,
            token=token,
          )
          if commit_hash is None:
            self.P("Could not get the commit hash. Assuming no new version.")
            result = False
          elif commit_hash != self.__git_commit_hash:        
            self.P(f"New git assets available: local hash {self.__git_commit_hash} differs from git {commit_hash} . Server reloading procedure will be initiated...")
            result = True
          # endif commit hash
        # endif can request git
      else:
        # no previous git info found so we assume we need to perform setup
        self.P("No previous local git info found. Assuming new repo version and initializing...")
        result = True
    return result

  # assets handling methods
  def __maybe_download_assets(self):
    """
      self.cfg_assets = {
        "url": "https://example.com/assets.zip",
        "operation": "download",
      }
      self.cfg_assets = {
        "url": "https://github.com/user/repo",
        "username": "username",
        "token": "token",
        "operation": "clone",
      }
      self.cfg_assets = {
        "url": "https://github.com/user/repo",
        "username": null,
        "token": null,
        "operation": "clone",
      }
      self.cfg_assets = {
        "url": "/path/to/local/dir",
        "operation": "download",
      }
      self.cfg_assets = {
        "url"=[["base64_encoded_file", "encoded_file_name"], ...],
        "operation": "decode",
      }
    """
    # handle assets url: download, extract, then copy, then delete
    relative_assets_path = self.os_path.join('downloaded_assets', self.plugin_id, 'assets')

    operation = None

    assets_path = None

    # check if assets is a dict or a string
    # and extract the url and operation
    if isinstance(self.cfg_assets, dict):
      dct_data = self.cfg_assets
      operation = dct_data.get("operation", "download")
      assets_path = dct_data.get("url", None)

    elif isinstance(self.cfg_assets, str):
      assets_path = self.cfg_assets
      operation = "download"

    if assets_path is None:
      return None
      # raise ValueError("No assets provided")

    # now download the assets there
    if operation == "clone":
      username = dct_data.get("username", None)
      token = dct_data.get("token", None)
      self.git_clone(
        repo_url=assets_path,
        repo_dir=relative_assets_path,
        target='output',
        user=username,
        token=token,
        pull_if_exists=True, # no need to pull if each time we delete the folder
      )
      # now we cache the commit hash
      commit_hash = self.git_get_local_commit_hash(
        repo_dir=self.os_path.join(self.get_output_folder(), relative_assets_path),
      )
      self.__git_commit_hash = commit_hash
      self.P("Finished cloning git repository. Current commit hash: {}".format(self.__git_commit_hash))
      #
    elif operation == "download":
      self.maybe_download(
        url=assets_path,
        fn=relative_assets_path,
        target='output'
      )
    elif operation == "decode":
      target_dir = self.os_path.join(self.get_output_folder(), relative_assets_path)
      for base64_encoded_file, encoded_file_name in assets_path:
        file_path = self.os_path.join(target_dir, encoded_file_name)
        os.makedirs(self.os_path.dirname(file_path), exist_ok=True)

        encoded_file = self.base64_to_str(base64_encoded_file, decompress=True)
        with open(file_path, 'w') as f:
          f.write(encoded_file)
        # endwith
      # endfor
    else:
      raise ValueError(f"Invalid operation {operation}")

    # now check if it is a zip file
    assets_path = self.os_path.join(self.get_output_folder(), relative_assets_path)
    if self.os_path.isfile(assets_path):
      if not assets_path.endswith('.zip'):
        os.rename(assets_path, assets_path + '.zip')
        assets_path += '.zip'

      relative_assets_path = self.os_path.join('downloaded_assets', self.plugin_id, 'unzipped')
      self.maybe_download(
        url=assets_path,
        fn=relative_assets_path,
        target='output',
        unzip=True
      )
      # remove zip file
      os.remove(assets_path)

    assets_path = self.os_path.join(self.get_output_folder(), relative_assets_path)
    return assets_path

  def __maybe_forced_reload(self):
    reload_interval = self.cfg_forced_reload_interval or 0
    if reload_interval > 0 and self.time() - self.webapp_reload_last_timestamp > reload_interval:
      self.P(f"Forced restart initiated due to `FORCED_RELOAD_INTERVAL` ({reload_interval} seconds)")
      self.__reload_server()
    return

  def __maybe_init_assets(self):
    if self.assets_initialized:
      self.__maybe_forced_reload()
      # check if new git assets are available
      new_repo_version = self.__check_new_repo_version()
      if new_repo_version:
        self.P("New git assets available. Reloading server...")
        self.__reload_server()
      return
    
    self.P("Initializing assets (`assets_initialized: {}`)".format(self.assets_initialized))

    # download/clone/create/unzip assets
    assets_path = self.__maybe_download_assets()

    # prepare environment variables
    self.__prepare_env(assets_path=assets_path)

    # initialize assets -- copy them to the temp directory
    self.initialize_assets(
      src_dir=assets_path,
      dst_dir=self.script_temp_dir,
      jinja_args=self.jinja_args
    )

    self.webapp_reload_last_timestamp = self.time()
    self.assets_initialized = True
    return

  
  
  # other plugin default methods
  
  def __setup_commands(self):
    self.setup_commands_started = [False] * len(self.get_setup_commands())
    self.setup_commands_finished = [False] * len(self.get_setup_commands())
    self.setup_commands_processes = [None] * len(self.get_setup_commands())
    self.setup_commands_start_time = [None] * len(self.get_setup_commands())

    self.start_commands_started = [False] * len(self.get_start_commands())
    self.start_commands_finished = [False] * len(self.get_start_commands())
    self.start_commands_processes = [None] * len(self.get_start_commands())
    self.start_commands_start_time = [None] * len(self.get_start_commands())

    self.logs = self.deque(maxlen=1000)
    self.dct_logs_reader = {}

    self.err_logs = self.deque(maxlen=1000)
    self.dct_err_logs_reader = {}

    self.P(f"Port: {self.port}")
    self.P(f"Setup commands: {self.get_setup_commands()}")
    self.P(f"Start commands: {self.get_start_commands()}")
    return


  def _on_close(self):
    self.__maybe_close_setup_commands()
    self.__maybe_close_start_commands()

    # close all log readers that are still running
    # this should not have any effect, as the logs should have been
    # closed during the teardown of the setup and start processes
    self.__maybe_read_and_stop_all_log_readers()

    # cleanup the temp directory
    shutil.rmtree(self.script_temp_dir)

    self.__deallocate_port()

    super(BaseWebAppPlugin, self)._on_close()
    return

  async def close_ngrok_listener(self):
    try:
      await self.ngrok_listener.close()
      self.P(f'Ngrok listener successfully closed.')
    except Exception as exc:
      self.P(f'Could not close Ngrok listener. Reason: {exc}')
    return

  def on_close(self):
    # This method is called by super(BaseWebAppPlugin, self)._on_close()
    super(BaseWebAppPlugin, self).on_close()
    if self.ngrok_listener is not None:
      self.P(f"Closing Ngrok listener...")
      asyncio.run(self.close_ngrok_listener())
    # endif ngrok listener opened
    return

  def _on_command(self, data, delta_logs=None, full_logs=None, start=None, reload=None, **kwargs):
    super(BaseWebAppPlugin, self)._on_command(data, **kwargs)

    if (isinstance(data, str) and data.upper() == 'DELTA_LOGS') or delta_logs:
      logs, err_logs = self.__get_delta_logs()
      self.add_payload_by_fields(
        command_params=data,
        logs=logs,
        err_logs=err_logs,
      )
    if (isinstance(data, str) and data.upper() == 'FULL_LOGS') or full_logs:
      # TODO: Implement full logs
      self.add_payload_by_fields(
        command_params=data,
        logs=[]
      )

    if (isinstance(data, str) and data.upper() == 'START') or start:
      self.can_run_start_commands = True
      self.P("Starting server")

    if (isinstance(data, str) and data.upper() == 'RELOAD') or reload:
      self.__reload_server()

    return

  # Exposed methods
  @property
  def port(self):
    if 'USED_PORTS' not in self.plugins_shmem:
      return None
    if self.str_unique_identification not in self.plugins_shmem['USED_PORTS']:
      return None
    port = self.plugins_shmem['USED_PORTS'][self.str_unique_identification]
    return port

  @property
  def jinja_args(self):
    return {}

  def get_setup_commands(self):
    try:
      return super(BaseWebAppPlugin, self).get_setup_commands() + self.cfg_setup_commands
    except AttributeError:
      return self.cfg_setup_commands

  def get_start_commands(self):
    try:
      return super(BaseWebAppPlugin, self).get_start_commands() + self.cfg_start_commands
    except AttributeError:
      return self.cfg_start_commands

  def initialize_assets(self, src_dir, dst_dir, jinja_args):
    """
    Initialize and copy assets, expanding any jinja templates.
    All files from the source directory are copied to the
    destination directory with the following exceptions:
      - are symbolic links are ignored
      - files named ending with .jinja are expanded as jinja templates,
        .jinja is removed from the filename and the result copied to
        the destination folder.
    This maintains the directory structure of the source folder.
    In case src_dir is None, only the .env files are written in the destination folder.

    Parameters
    ----------
    src_dir: str or None, path to the source directory
    dst_dir: str, path to the destination directory
    jinja_args: dict, jinja keys to use while expanding the templates

    Returns
    -------
    None
    """

    # now copy the assets to the destination
    self.P(f'Copying assets from {src_dir} to {dst_dir} with keys {jinja_args}')

    if src_dir is not None:
      env = Environment(loader=FileSystemLoader('.'))
      # Walk through the source directory.
      for root, _, files in os.walk(src_dir):
        for file in files:
          src_file_path = self.os_path.join(root, file)
          dst_file_path = self.os_path.join(
            dst_dir,
            self.os_path.relpath(src_file_path, src_dir)
          )

          # If we have a symlink don't do anything.
          if self.os_path.islink(src_file_path):
            continue

          # Make sure the destination directory exists.
          os.makedirs(self.os_path.dirname(dst_file_path), exist_ok=True)

          is_jinja_template = False
          if src_file_path.endswith(('.jinja')):
            dst_file_path = dst_file_path[:-len('.jinja')]
            is_jinja_template = True
          if src_file_path.endswith(('.j2')):
            dst_file_path = dst_file_path[:-len('.j2')]
            is_jinja_template = True

          # If this file is a jinja template render it to a file with the .jinja suffix removed.
          # Otherwise just copy the file to the destination directory.
          if is_jinja_template:
            template = env.get_template(src_file_path)
            rendered_content = template.render(jinja_args)
            with open(dst_file_path, 'w') as f:
              f.write(rendered_content)
          else:
            shutil.copy2(src_file_path, dst_file_path)
          # endif is jinja template
        # endfor all files
      # endfor os.walk
    # endif src_dir is not None

    # write .env file in the target directory
    # environment variables are passed in subprocess.Popen, so this is not needed
    # but it's useful for debugging
    with open(self.os_path.join(self.script_temp_dir, '.env'), 'w') as f:
      for key, value in self.prepared_env.items():
        f.write(f"{key}={value}\n")

    with open(self.os_path.join(self.script_temp_dir, '.start_env_used'), 'w') as f:
      for key, value in self.prepared_env.items():
        f.write(f"{key}={value}\n")

    with open(self.os_path.join(self.script_temp_dir, '.setup_env_used'), 'w') as f:
      for key, value in self.base_env.items():
        f.write(f"{key}={value}\n")

    # now cleanup the download folder if it exists
    downloaded_assets_path = self.os_path.join(self.get_output_folder(), 'downloaded_assets', self.plugin_id)
    if self.os_path.exists(downloaded_assets_path):
      shutil.rmtree(downloaded_assets_path)

    self.P("Assets copied successfully")

    return

  def __maybe_ngrok_ping(self):
    # Check if the Ngrok API is used.
    if not self.cfg_ngrok_use_api:
      return
    # Check if the listener is available.
    if self.ngrok_listener is None:
      return
    # Check if the listener has a URL.
    # In case a Ngrok edge label or domain is provided no URL will be available since the user should already have it.
    if self.ngrok_listener.url() is None:
      return
    if self.__last_ngrok_url_ping_ts is None or self.time() - self.__last_ngrok_url_ping_ts >= self.cfg_ngrok_url_ping_interval:
      self.__last_ngrok_url_ping_ts = self.time()
      self.add_payload_by_fields(
        ngrok_url=self.ngrok_listener.url(),
      )
    # endif last ngrok url ping
    return

  def __maybe_setup_commands(self):
    if self.__commands_ready:
      return
    self.__setup_commands()
    self.__commands_ready = True
    return

  def _process(self):  # Check: _process as opposed to process
    self.__maybe_setup_commands()

    self.__maybe_init_assets()  # Check: auto-updates point here?

    self.maybe_init_ngrok()

    self.__maybe_run_all_setup_commands()

    self.maybe_start_ngrok()

    self.__maybe_run_all_start_commands()

    self.__maybe_print_all_logs() # Check: must review 

    super(BaseWebAppPlugin, self)._process() # Check: why is this required
    self.__maybe_ngrok_ping()
    return



from naeural_core.business.base import BasePluginExecutor as BasePlugin
from naeural_core.business.mixins_libs.nlp_agent_mixin import _NlpAgentMixin, NLP_AGENT_MIXIN_CONFIG

__VER__ = '0.1.0.0'

_CONFIG = {
  # mandatory area
  **BasePlugin.CONFIG,
  **NLP_AGENT_MIXIN_CONFIG,

  # our overwritten props
  'AI_ENGINE': "llm",

  'VALIDATION_RULES': {
    **BasePlugin.CONFIG['VALIDATION_RULES'],
    **NLP_AGENT_MIXIN_CONFIG['VALIDATION_RULES'],
  },
}


class LlmAgentPlugin(BasePlugin, _NlpAgentMixin):
  CONFIG = _CONFIG

  def _process(self):
    # we always receive input from the upstream due to the fact that _process
    # is called only when we have input based on ALLOW_EMPTY_INPUTS=False (from NLP_AGENT_MIXIN_CONFIG)
    data = self.dataapi_struct_data()
    self.P(f"Received request:\n{self.json_dumps(self.shorten_str(data), indent=2)}")
    inferences = self.dataapi_struct_data_inferences()
    return self.compute_response(data, inferences)



from naeural_core.serving.base.base_llm_serving import BaseLlmServing as BaseServingProcess
from transformers import AutoTokenizer, AutoModel
import re

from docarray import BaseDoc, DocList
from docarray.typing import NdArray
from vectordb import HNSWVectorDB


"""
  TODO:
  - integrate vectordb library
  - add segmentation of the context
  - support multiple sets of context(maybe a dictionary of format {key: list[doc1, doc2, ...]})
  - add context to a single set
  - change context for a single set
  - reset all sets of context
  
"""


__VER__ = '0.1.0.0'
EMBEDDING_SIZE = 512
MAX_SEGMENT_SIZE = 300
MAX_SEGMENT_OVERLAP = 30
WORD_FIND_REGEX = r'\b(?:[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}|[a-zA-Z]+(?:\'[a-z]+)?|[0-9]+(?:\.[0-9]+)?|[^\s\w])\b'
DEFAULT_NUMBER_OF_RESULTS = 10

_CONFIG = {
  **BaseServingProcess.CONFIG,

  'MAX_BATCH_SIZE': 32,
  'MAX_EMB_SIZE': EMBEDDING_SIZE,
  'RUNS_ON_EMPTY_INPUT': True,

  'VALIDATION_RULES': {
    **BaseServingProcess.CONFIG['VALIDATION_RULES'],
  },

}


class DocEmbCt:
  REQUEST_TYPE = 'REQUEST_TYPE'
  REQUEST_ID = 'REQUEST_ID'
  REQUEST_PARAMS = 'REQUEST_PARAMS'

  QUERY = 'QUERY'
  ADD_DOC = 'ADD_DOC'
  LIST_CONTEXT = 'LIST_CONTEXT'

  BAD_REQUEST = 'BAD_REQUEST'
  ERROR_MESSAGE = 'ERROR_MESSAGE'
  DEFAULT_REQUEST_TYPE = QUERY
  REQUEST_TYPES = [QUERY, ADD_DOC, LIST_CONTEXT]

  DOC_KEY = 'doc'
  DOCS_KEY = 'docs'
  URL_KEY = 'url'
  AVAILABLE_DOCS = ['doc', 'docs']
# endclass


class NaeuralDoc(BaseDoc):
  # TODO: encrypt the text for db and decrypt it when needed.
  text: str = ''
  embedding: NdArray[EMBEDDING_SIZE]
  idx: int = -1
# endclass


class DocSplitter:
  """
  Class for splitting one document or a list of documents into segments.
  """
  def __init__(
      self, max_segment_size: int = MAX_SEGMENT_SIZE,
      max_segment_overlap: int = MAX_SEGMENT_OVERLAP,
  ):
    """
    Constructor for the DocSplitter class.

    Parameters
    ----------
    max_segment_size : int - the maximum size of a segment in words
    max_segment_overlap : int - the maximum overlap between segments in words
    """
    self.__max_segment_size = max_segment_size
    self.__max_segment_overlap = max_segment_overlap
    return

  def document_atomizing(self, document: str):
    """
    Atomize a document into words.

    Parameters
    ----------
    document : str - the document to be atomized

    Returns
    -------
    list[str] - the list of words
    """
    return re.findall(WORD_FIND_REGEX, document)

  def compute_best_overlap(self, text_size: int):
    """
    Compute the best overlap for the segments, while considering the text size and the maximum segment size.

    Parameters
    ----------
    text_size : int - the size of the text

    Returns
    -------
    int - the best overlap for the segments
    """
    # In case the text size is smaller than the maximum segment size, we do not need any overlap.
    if text_size <= self.__max_segment_size:
      return 0
    # We want the last segment to be as long as possible.
    best_overlap, best_last_segment_length = 0, 0
    # We cannot have an overlap larger than the maximum segment overlap.
    max_overlap = min(self.__max_segment_overlap, self.__max_segment_size - 1)
    min_overlap = int(max_overlap // 2)
    # The overlap between the segments helps us keep as much information as possible.
    # Thus, we start from half of the maximum overlap and go up to the maximum overlap.
    for overlap in range(min_overlap, max_overlap + 1):
      last_segment_length = (text_size - self.__max_segment_size) % (self.__max_segment_size - overlap)
      if last_segment_length > best_last_segment_length:
        best_overlap, best_last_segment_length = overlap, last_segment_length
      # endif last segment length is better
    # endfor each overlap
    return best_overlap

  def split_document(self, document: str):
    """
    Split a document into segments.

    Parameters
    ----------
    document : str - the document to be split

    Returns
    -------
    list[str] - the list of segments
    """
    # Break the document in words.
    words = self.document_atomizing(document)
    # Compute the best overlap for the segments.
    overlap = self.compute_best_overlap(text_size=len(words))
    increment_step = max(1, self.__max_segment_size - overlap)
    # Split the document into segments.
    segments = [
      ' '.join(words[i:i + self.__max_segment_size])
      for i in range(0, max(1, len(words) - overlap), increment_step)
    ]
    return segments

  def split_documents(self, documents: list[str]):
    """
    Split a list of documents into segments.
    Each document will be split into segments on its own, but the segmentations will be concatenated.

    Parameters
    ----------
    documents : list[str] - the list of documents to be split

    Returns
    -------
    list[str] - the list of segments
    """
    segmentations = [self.split_document(doc) for doc in documents]
    return sum(segmentations, [])
# endclass DocSplitter


class BaseDocEmbServing(BaseServingProcess):
  CONFIG = _CONFIG
  def __init__(self, **kwargs):
    super(BaseDocEmbServing, self).__init__(**kwargs)
    self.__dbs = {}
    self.__doc_splitter = DocSplitter()
    return

  def __context_identifier(self, context):
    return 'default' if context is None else f'context_{context}'

  def __db_cache_workspace(self, context):
    return self.os_path.join(self.get_models_folder(), 'vectordb', self.cfg_model_name, context)

  def __backup_contexts(self):
    """
    Backup the contexts to ensure their persistence.
    """
    self.persistence_serialization_save(
      obj={
        'contexts': list(self.__dbs.keys())
      }
    )
    return

  def __maybe_load_backup(self):
    """
    In case of persisted contexts, load them.
    """
    saved_data = self.persistence_serialization_load()
    if saved_data is not None:
      contexts = saved_data.get('contexts', [])
      for context in contexts:
        if context not in self.__dbs:
          self.__dbs[context] = HNSWVectorDB[NaeuralDoc](workspace=self.__db_cache_workspace(context))
        # endif sanity check in case of db already loaded
      # endfor each context
    # endif saved data available
    return

  def on_init(self):
    super(BaseDocEmbServing, self).on_init()
    self.__maybe_load_backup()
    return

  def _setup_llm(self):
    # just override this method as the base class has a virtual method that raises an exception
    return

  def _get_device_map(self):
    return self.device

  def load_tokenizer(self, model_id, cache_dir, token):
    self.tokenizer = AutoTokenizer.from_pretrained(
      model_id,
      cache_dir=self.cache_dir,
      use_auth_token=self.hf_token
    )
    return

  def load_pretrained_model(self, model_id, **kwargs):
    return AutoModel.from_pretrained(model_id, **kwargs)

  def _warmup(self):
    warmup_context = [
      "The Tesla Cybertruck is a battery electric pickup truck built by Tesla, Inc. since 2023.[6] Introduced as a "
      "concept vehicle in November 2019, it has a body design reminiscent of low-polygon modelling, consisting of flat "
      "stainless steel sheet panels.\nTesla initially planned to produce the vehicle in 2021, but it entered "
      "production in 2023 and was first delivered to customers in November. Three models are offered: a tri-motor "
      "all-wheel drive (AWD) \"Cyberbeast\", a dual-motor AWD model, and a rear-wheel drive (RWD) model, with EPA "
      "range estimates of 250–340 miles (400–550 km), varying by model.\nAs of December 2023, the Cybertruck is "
      "available only in North America.",

      "Am facut acest chec pufos cu cacao de atata ori si pentru atat de multe ocazii, incat cred ca-l pot face cu "
      "ochii inchisi. Checul este unul din deserturile clasice romanesti. Il faceau bunicile noastre, mamele noastre "
      "si acum este randul nostru sa ducem reteta mai departe. Este atat de iubit si de popular incat tuturor le "
      "place. Mama este una dintre marile iubitoarele acestui chec, la fel ca mine, de altfel. Alaturi de reteta de "
      "cozonac, checul este desertul pe care il facea cel mai des. Ni l-a facut toata copilaria si imi amintesc cu "
      "drag si nostalgie de feliile groase de chec presarate din abundenta cu zahar pudra. Era minunat pentru micul "
      "dejun, dar si ca gustare, alaturi de un pahar cu lapte sau de o cafea. Il manacam imediat si rar ne mai ramanea "
      "si a doua zi.\nReteta aceasta de chec pufos cu cacao este putin diferita de cea pe care o facea mama. Am "
      "modificat-o in asa fel incat sa fie usor de facut si sa reduc la minim riscul de a da gres. Cel mai important "
      "lucru atunci cand faceti aceasta reteta este sa bateti cat mai bine albusurile. Trebuie sa incorporati cat mai "
      "mult aer in ele. Pentru asta puteti folosi un stand-mixer sau pur si simplu un mixer manual. Puteti incerca si "
      "cu un tel, insa va dura considerabil mai mult timp. Aveti grija cand separati albusurile! Nicio picatura de "
      "galbenus nu trebuie sa ajunga in ele. La fel, nicio picatura de grasime, altfel nu se vor bate cum trebuie. Si "
      "bolul trebuie sa fie bine spalat si degresat cu putina zeama de lamaie sau otet.Evitati sa folositi boluri din "
      "plastic pentru ca nu se vor curata la fel de bine."
    ]
    warmup1 = warmup_context[:1]
    warmup4 = warmup_context + warmup_context
    self.P(f'Warming up with {len(warmup1)} texts')
    self.embed_texts(warmup_context[:1])
    self.P(f'Warming up with {len(warmup4)} texts')
    self.embed_texts(warmup_context + warmup_context)
    self.P(f'Warmup done')
    return

  """PREPROCESS OF REQUESTS"""
  if True:
    def processed_bad_request(self, msg, request_id=None, predict_kwargs=None):
      return {
        DocEmbCt.REQUEST_ID: request_id,
        DocEmbCt.REQUEST_TYPE: DocEmbCt.BAD_REQUEST,
        DocEmbCt.REQUEST_PARAMS: {},
        DocEmbCt.ERROR_MESSAGE: msg,
        'PREDICT_KWARGS': predict_kwargs or {},
      }

    """VALIDATION OF REQUESTS"""
    if True:
      def doc_embedding_valid_docs(self, docs_value):
        is_bad_request, processed_request_params, err_msg = False, {}, ""
        if not isinstance(docs_value, list) or not all([isinstance(x, str) for x in docs_value]):
          additional = ""
          if isinstance(docs_value, list):
            non_str_types = [type(x) for x in docs_value if not isinstance(x, str)]
            non_str_types = list(set(non_str_types))
            additional = f" containing non string types: {non_str_types}"
          # endif list, but not all strings
          err_msg = (f"Error! For ADD_DOC request `{DocEmbCt.REQUEST_PARAMS}` the `docs` key must be a list of strings."
                     f"Received {type(docs_value)}{additional}.")
          is_bad_request = True
        else:
          processed_request_params = {'docs': docs_value}
        # endif valid docs_value
        return is_bad_request, processed_request_params, err_msg

      def doc_embedding_valid_doc(self, doc_value):
        is_bad_request, processed_request_params, err_msg = False, {}, ""
        if not isinstance(doc_value, str):
          err_msg = (f"Error! For ADD_DOC request `{DocEmbCt.REQUEST_PARAMS}` the `doc` key must be a string."
                     f"Received {type(doc_value)}.")
          is_bad_request = True
        else:
          processed_request_params = {'docs': [doc_value]}
        # endif valid doc_value
        return is_bad_request, processed_request_params, err_msg

      def doc_embedding_valid_url(self, url_value):
        is_bad_request, processed_request_params, err_msg = False, {}, ""
        err_msg = "Error! The `url` key is not available for the moment."
        is_bad_request = True
        return is_bad_request, processed_request_params, err_msg

      def doc_embedding_validate_request_params(self, request_type, request_params):
        """
        Method for validating the request parameters.

        Parameters
        ----------
        request_type : str - the request type
        request_params : dict - the request parameters

        Returns
        -------

        is_bad_request : bool - whether the request is bad
        processed_request_params : dict - the processed request parameters
        err_msg : str - the error message if any
        """
        is_bad_request, processed_request_params, err_msg = False, {}, ""
        # Normalize the keys to lowercase.
        lowercase_params = {k.lower(): v for k, v in request_params.items()}

        if request_type == DocEmbCt.LIST_CONTEXT:
          # No additional parameters are needed
          pass
        elif request_type == DocEmbCt.QUERY:
          query_value = lowercase_params.get('query', None)
          if query_value is None or not isinstance(query_value, str):
            err_msg = (f"Error! `{DocEmbCt.REQUEST_PARAMS}` must contain a 'query' key with a string value. "
                       f"Received {type(query_value)}.")
            is_bad_request = True
          else:
            processed_request_params = {
              'query': query_value,
              'k': lowercase_params.get('k', 10)
            }
          # endif query not in request params
        elif request_type == DocEmbCt.ADD_DOC:
          doc_value = lowercase_params.get('doc', None)
          docs_value = lowercase_params.get('docs', None)
          url_value = lowercase_params.get('url', None)

          if doc_value is None and docs_value is None and url_value is None:
            err_msg = (f"Error! For ADD_DOC request `{DocEmbCt.REQUEST_PARAMS}` must contain at least "
                       f"one of {DocEmbCt.AVAILABLE_DOCS} keys. Received {list(lowercase_params.keys())}.")
            is_bad_request = True
          else:
            # At least one is present
            if docs_value is not None:
              is_bad_request, processed_request_params, err_msg = self.doc_embedding_valid_docs(docs_value)
            elif doc_value is not None:
              is_bad_request, processed_request_params, err_msg = self.doc_embedding_valid_doc(doc_value)
            elif url_value is not None:
              is_bad_request, processed_request_params, err_msg = self.doc_embedding_valid_url(url_value)
          # endif doc, docs and url are None
        else:
          # This should not happen. We already checked the request type and this is only a sanity check.
          err_msg = f"Error! `{DocEmbCt.REQUEST_TYPE}` value must be one of {DocEmbCt.REQUEST_TYPES}. Received {request_type}."
          is_bad_request = True
        # endif request_params checks
        if not is_bad_request:
          processed_request_params['context'] = lowercase_params.get('context', None)
        # endif not bad request
        return is_bad_request, processed_request_params, err_msg
    """END VALIDATION OF REQUESTS"""

    def get_additional_metadata(self):
      return {
        'MODEL_NAME': self.cfg_model_name,
        'EMBEDDING_SIZE': EMBEDDING_SIZE,
        'MAX_SEGMENT_SIZE': MAX_SEGMENT_SIZE,
        'CONTEXTS': list(self.__dbs.keys())
      }

    def _pre_process(self, inputs):
      """
      Preprocess the inputs for the prediction.
      The inputs should have the following format:
      {
        'DATA': [
          {
            # Will add doc to the default context
            'REQUEST_ID': 'request_id_1',
            'REQUEST_TYPE': 'ADD_DOC',
            'REQUEST_PARAMS': {
              'doc': 'text1',
            }
          },
          {
            # Will add docs to the default context
            'REQUEST_ID': 'request_id_2',
            'REQUEST_TYPE': 'ADD_DOC',
            'REQUEST_PARAMS': {
              'docs': ['text2', 'text3'],
            }
          },
          {
            # This is unavailable for the moment.
            # Will add the content from https://www.example.com to 'context1'
            'REQUEST_ID': 'request_id_3',
            'REQUEST_TYPE': 'ADD_DOC',
            'REQUEST_PARAMS': {
              'url': 'https://www.example.com',
              'context': 'context1',
            }
          },
          {
            # Will compute the closest 10 documents to 'query1' in the default context
            'REQUEST_ID': 'request_id_4',
            'REQUEST_TYPE': 'QUERY',
            'REQUEST_PARAMS': {
              'query': 'query1',
              'k': 10,
            }
          }
          ...
        ],
        'SERVING_PARAMS': [
          {
            'param1': 'value1',
            'param2': 'value2',
          },
          {},
          ...
        ]
      }, where SERVING_PARAMS is optional and contains additional parameters for the prediction.
      The requests can be of the following types:

      - ADD_DOC:
        The request params must contain either a 'doc' or a 'docs' key.
        The 'doc' key must have a string value representing the document to be added to the specified context.
        The 'docs' key must have a list of strings representing the documents to be added to the specified context.
        The 'url' key is unavailable for the moment.
        If both keys are present, the 'docs' key will be used.
        If no context is specified through the 'context' key in the "REQUEST_PARAMS" dict,
        the default context will be used.

      - QUERY:
        The request params must contain a 'query' key with a string value representing the query to be solved.
        The 'k' key is optional and must have an integer value representing the number of closest documents to be returned.
        If the 'k' key is not present, the default value of 10 will be used.
        Same as the 'ADD_DOC' request, the 'context' key is used to specify the context to be used.

      - LIST_CONTEXT:
        No additional parameters are needed.
        This request will return the list of available contexts.

      Parameters
      ----------
      inputs : dict - the inputs for the prediction

      Returns
      -------

      processed_requests : list[dict] - the processed requests

      """
      lst_inputs = inputs.get('DATA', [])
      serving_params = inputs.get('SERVING_PARAMS', [])
      if len(lst_inputs) == 0:
        return []
      # endif no inputs

      processed_requests = []
      for i, inp in enumerate(lst_inputs):
        is_bad_request = False
        msg = ""
        # Check if the input is a dictionary
        if not isinstance(inp, dict):
          msg = f"Error! Input {i} must be a dict. Received {type(inp)}!"
          self.P(msg)
          processed_requests.append(self.processed_bad_request(msg))
          continue
        # endif input is a string

        predict_kwargs = serving_params[i] if i < len(serving_params) else {}
        normalized_input = {k.upper(): v for k, v in inp.items()}
        request_id = normalized_input.get(DocEmbCt.REQUEST_ID, None)
        if request_id is None:
          msg = f"Warning! Request {i} must have a request id specified in `{DocEmbCt.REQUEST_ID}`."
          self.P(msg)
        # endif request_id provided

        # Check request type
        request_type = normalized_input.get(DocEmbCt.REQUEST_TYPE, DocEmbCt.DEFAULT_REQUEST_TYPE)
        if request_type not in DocEmbCt.REQUEST_TYPES:
          msg = f"Error! `{DocEmbCt.REQUEST_TYPE}` value must be one of {DocEmbCt.REQUEST_TYPES}. Received {request_type}."
          self.P(msg)
          processed_requests.append(self.processed_bad_request(msg, request_id=request_id))
          continue
        # endif request type is not valid

        # Check request params
        request_params = normalized_input.get(DocEmbCt.REQUEST_PARAMS, {})
        if not isinstance(request_params, dict) and request_type != DocEmbCt.LIST_CONTEXT:
          msg = f"Error! `{DocEmbCt.REQUEST_PARAMS}` value must be a dict. Received {type(request_params)}!"
          self.P(msg)
          processed_requests.append(self.processed_bad_request(msg, request_id=request_id))
          continue
        # endif request params is not a dict

        # Validate the request params
        is_bad_request, processed_request_params, msg = self.doc_embedding_validate_request_params(
          request_type=request_type, request_params=request_params
        )

        processed_requests.append({
          DocEmbCt.REQUEST_ID: request_id,
          DocEmbCt.REQUEST_TYPE: request_type if not is_bad_request else DocEmbCt.BAD_REQUEST,
          DocEmbCt.REQUEST_PARAMS: processed_request_params if not is_bad_request else {},
          DocEmbCt.ERROR_MESSAGE: msg,
          'PREDICT_KWARGS': predict_kwargs
        })
      # endfor each input
      return processed_requests
  """END PREPROCESS OF REQUESTS"""

  """PROCESSING OF REQUESTS"""
  if True:
    def pooling(self, last_hidden_states, attention_mask):
      """
      Pool the last hidden states using the attention mask.
      Parameters
      ----------
      last_hidden_states : torch.Tensor (batch_size, seq_len, hidden_size) with the last hidden states
      attention_mask : torch.Tensor (batch_size, seq_len) with 0s for padding and 1s for real tokens

      Returns
      -------
      torch.Tensor (batch_size, hidden_size) with the pooled embeddings
      """
      return (self.th.sum(last_hidden_states * attention_mask.unsqueeze(-1), dim=1) /
              self.th.sum(attention_mask, dim=1, keepdim=True))

    def embed_texts(self, texts):
      """
      Embed the texts using the model.
      Parameters
      ----------
      texts : str or list[str] - the text or the list of texts to be embedded

      Returns
      -------

      """
      if not isinstance(texts, list):
        texts = [texts]
      # endif texts is not a list
      if self.cfg_max_batch_size is not None and len(texts) > self.cfg_max_batch_size:
        batches = [texts[i:i + self.cfg_max_batch_size] for i in range(0, len(texts), self.cfg_max_batch_size)]
      else:
        batches = [texts]
      # endif more texts than max batch size
      embeddings = []
      for batch in batches:
        with self.th.no_grad():
          input_dict = self.tokenizer(
            batch, max_length=self.cfg_max_emb_size, padding=True, truncation=True, return_tensors='pt'
          )
          input_dict = {k: v.to(self.device) for k, v in input_dict.items()}
          outputs = self.model(**input_dict)
        # endwith no grad
        current_embeddings = self.pooling(outputs.last_hidden_state, input_dict['attention_mask'])
        current_embeddings = self.th.nn.functional.normalize(current_embeddings, p=2, dim=1)
        embeddings.append(current_embeddings.to('cpu'))
        self.th_utils.clear_cache()
      # endfor each batch
      return self.th.cat(embeddings, dim=0)

    def __add_docs(self, docs, context: str = None):
      """
      Add the documents to the context.
      Parameters
      ----------
      docs : list[str] - the list of documents
      context : str - the context name
      """
      context = self.__context_identifier(context)
      # endif context is None
      if context not in self.__dbs:
        self.__dbs[context] = HNSWVectorDB[NaeuralDoc](workspace=self.__db_cache_workspace(context))
        self.__backup_contexts()
      # endif context not in dbs
      segments = self.__doc_splitter.split_documents(docs)
      segments_embeddings = self.embed_texts(segments)
      curr_size = self.__dbs[context].num_docs()['num_docs']
      lst_docs = [
        NaeuralDoc(text=segment, embedding=emb, idx=curr_size + i)
        for i, (segment, emb) in enumerate(zip(segments, segments_embeddings))
      ]
      # TODO: maybe check for duplicates
      self.__dbs[context].index(inputs=DocList[NaeuralDoc](lst_docs))
      return

    def get_result_dict(self, request_id, docs=None, query=None, context_list=None, error_message=None, **kwargs):
      """
      Get the result dictionary.
      Parameters
      ----------
      request_id : str - the request id
      docs : list[str] - the document list
      query : str - the query
      context_list : list[str] - the list of available contexts
      error_message : str - the error message, in case of an error
      kwargs : dict - additional parameters

      Returns
      -------
      dict - the result dictionary
      """
      uppercase_kwargs = {k.upper(): v for k, v in kwargs.items()}
      return {
        DocEmbCt.REQUEST_ID: request_id,
        'DOCS': docs,
        DocEmbCt.QUERY: query,
        'CONTEXT_LIST': context_list,
        'MODEL_NAME': self.cfg_model_name,
        DocEmbCt.ERROR_MESSAGE: error_message,
        **uppercase_kwargs
      }

    def _predict(self, preprocessed_requests):
      """
      Perform the prediction using the preprocessed requests.
      For details about the requests see the `_pre_process` method.
      Parameters
      ----------
      preprocessed_requests : list[dict] - the preprocessed requests
        - each dict must have the following keys:
          - REQUEST_ID : str - the request id
          - REQUEST_TYPE : str - the request type: QUERY, ADD_DOC, LIST_CONTEXT
          - REQUEST_PARAMS : dict - the request parameters - can vary depending on the request type
        - each dict can have the following keys(they are optional):
          - PREDICT_KWARGS(not used for the moment) : dict - the prediction kwargs,
          additional parameters for the prediction

      Returns
      -------
      list[dict] - the predictions for each query or context query
        - each dict must have the following keys
          - REQUEST_ID : str - the request id
          - DOCS : list[str] - the requested documents, empty in case of ADD_DOC or LIST_CONTEXT
          - QUERY : str - the query, None if not a query
          - CONTEXT_LIST : list[str] - the list of available contexts in case of LIST_CONTEXT or None
          - MODEL_NAME : str - the model name
          - ERROR_MESSAGE : str - the error message, if any
          - additional keys can be added
      """
      results = []
      for i, req in enumerate(preprocessed_requests):
        req_id = req[DocEmbCt.REQUEST_ID]
        req_type = req[DocEmbCt.REQUEST_TYPE]
        if req_type == DocEmbCt.LIST_CONTEXT:
          results.append(
            self.get_result_dict(request_id=req_id, context_list=list(self.__dbs.keys()))
          )
        elif req_type == DocEmbCt.ADD_DOC:
          req_params = req[DocEmbCt.REQUEST_PARAMS]
          doc = req_params.get('doc', None)
          context = req_params.get('context', None)
          docs = req_params.get('docs', []) if doc is None else [doc]
          # TODO: with this implementation a context may be influenced by multiple sets of users.
          #  This can lead to a context that is not representative for any of the users.
          self.__add_docs(docs, context)
          results.append(self.get_result_dict(request_id=req_id))
        elif req_type == DocEmbCt.QUERY:
          # TODO: maybe support the following query:
          #  query + temporary context => the context will not be saved, but will be
          #  segmented and used for the query.
          req_params = req[DocEmbCt.REQUEST_PARAMS]
          query = req_params['query']
          context = req_params.get('context', None)
          context = self.__context_identifier(context)
          k = req_params.get('k', DEFAULT_NUMBER_OF_RESULTS)
          # In case the context is not available, return an error message.
          if context not in self.__dbs:
            results.append(
              self.get_result_dict(request_id=req_id, error_message=f"Error! Context {context} not found.")
            )
            continue
          # endif context not in dbs
          # Embed the query.
          query_embedding = self.embed_texts(query)
          query_doc = NaeuralDoc(text=query, embedding=query_embedding, idx=-1)
          # Search for the closest documents.
          search_results = self.__dbs[context].search(inputs=DocList[NaeuralDoc]([query_doc]), limit=k)
          # Sort the results by the index.
          result_texts = [
            res.text for res in sorted(search_results, key=lambda x: x.idx)
          ]
          results.append(
            self.get_result_dict(request_id=req_id, docs=result_texts, query=query)
          )
        elif req_type == DocEmbCt.BAD_REQUEST:
          err_msg = req[DocEmbCt.ERROR_MESSAGE]
          results.append(
            self.get_result_dict(request_id=req_id, error_message=err_msg)
          )
        # endif request type
      # endfor each preprocessed request
      return results
  """END PROCESSING OF REQUESTS"""

  def _post_process(self, preds_batch):
    return preds_batch
# endclass BaseDocEmbServing




"""
TODO:


  - Solve the issue for set-contention in the chain storage when two nodes try to set the same key at the same time
    - (better) implement a lock mechanism for the chain storage when setting a key value that 
      will allow multiple nodes to compete for the set-operation and thus load-balance it
      OR
    - implement set-value moving TOKEN via the network
  

  - peer-to-peer managed:
    - node 1 registers a app_id X via a list of peers that includes node 1, node 2 and node 3
    - node 2 sets value by:
      - key
      - value 
      - app_id X
    - node 2 broadcasts the set opration to all app_id X peers (not all peers)
    - node 2 waits for confirmations from at least half of the app_id peers
    - if node 4 tries to set key in app_id X, it will be rejected
    
  


"""


from naeural_core.business.base.network_processor import NetworkProcessorPlugin 

_CONFIG = {
  **NetworkProcessorPlugin.CONFIG,
  
  'ALLOW_EMPTY_INPUTS' : True,  
  "ACCEPT_SELF" : False,
  
  "FULL_DEBUG_PAYLOADS" : False,
  "CHAIN_STORE_DEBUG" : False, # main debug flag
  
  
  "MIN_CONFIRMATIONS" : 1,
  
  "CHAIN_PEERS_REFRESH_INTERVAL" : 60,

  'VALIDATION_RULES' : { 
    **NetworkProcessorPlugin.CONFIG['VALIDATION_RULES'],
  },  
}

__VER__ = '0.8.1'

class ChainStoreBasePlugin(NetworkProcessorPlugin):
  CONFIG = _CONFIG
  
  CS_STORE = "SSTORE"
  CS_CONFIRM = "SCONFIRM"
  CS_DATA = "CHAIN_STORE_DATA"
  CS_CONFIRM_BY = "confirm_by"
  CS_CONFIRM_BY_ADDR = "confirm_by_addr"
  CS_CONFIRMATIONS = "confirms"
  CS_MIN_CONFIRMATIONS = "min_confirms"
  CS_OP = "op"
  CS_KEY = "key"
  CS_VALUE = "value"
  CS_OWNER = "owner"
  CS_STORAGE_MEM = "__chain_storage" # shared memory key
  CS_GETTER = "__chain_storage_get"
  CS_SETTER = "__chain_storage_set"
  
  
  
  
  def on_init(self):
    super().on_init() # not mandatory anymore?
    
    self.P(" === ChainStoreBasicPlugin INIT")
    
    self.__chainstore_identity = "CS_MSG_{}".format(self.uuid(7))
    
    self.__ops = self.deque()
    
    try:
      self.__chain_storage = self.cacheapi_load_pickle(default={}, verbose=True)
    except Exception as e:
      self.P(f" === Chain storage could not be loaded: {e}")
      self.__chain_storage = {}
      

    ## DEBUG ONLY:
    if self.CS_STORAGE_MEM in self.plugins_shmem:
      self.P(" === Chain storage already exists", color="r")
      self.__chain_storage = self.plugins_shmem[self.CS_STORAGE_MEM]
    ## END DEBUG ONLY
    
    self.plugins_shmem[self.CS_STORAGE_MEM] = self.__chain_storage
    self.plugins_shmem[self.CS_GETTER] = self._get_value
    self.plugins_shmem[self.CS_SETTER] = self._set_value
    
    self.__last_chain_peers_refresh = 0
    self.__chain_peers = []
    self.__maybe_refresh_chain_peers()
    return
  
  
  def __debug_dump_chain_storage(self):
    if self.cfg_chain_store_debug:
      self.P(" === Chain storage dump:\n{}".format(self.json_dumps(self.__chain_storage, indent=2)))
    return
  
  
  def __maybe_refresh_chain_peers(self):
    if (self.time() - self.__last_chain_peers_refresh) > self.cfg_chain_peers_refresh_interval:
      _chain_peers = self.bc.get_whitelist(with_prefix=True)
      # now check and preserve only online peers
      self.__chain_peers = [
        peer for peer in _chain_peers if self.netmon.network_node_is_online(peer)
      ]
      self.__last_chain_peers_refresh = self.time()
    return
  
  
  def __send_data_to_chain_peers(self, data):
    self.send_encrypted_payload(node_addr=self.__chain_peers, **data)
    return
  
  
  def __get_min_peer_confirmations(self):
    if self.cfg_min_confirmations is not None and self.cfg_min_confirmations > 0:
      return self.cfg_min_confirmations    
    return len(self.__chain_peers) // 2 + 1
  
  
  def __save_chain_storage(self):
    self.cacheapi_save_pickle(self.__chain_storage, verbose=True)
    self.__last_chain_storage_save = self.time()
    return
  
  ## START setter-getter methods

  def __get_key_value(self, key):
    return self.__chain_storage.get(key, {}).get(self.CS_VALUE, None)


  def __get_key_owner(self, key):
    return self.__chain_storage.get(key, {}).get(self.CS_OWNER, None)


  def __get_key_confirmations(self, key):
    return self.__chain_storage.get(key, {}).get(self.CS_CONFIRMATIONS, 0)
  
  def __get_key_min_confirmations(self, key):
    return self.__chain_storage.get(key, {}).get(self.CS_MIN_CONFIRMATIONS, 0)


  def __reset_confirmations(self, key):
    self.__chain_storage[key][self.CS_CONFIRMATIONS] = 0
    self.__chain_storage[key][self.CS_MIN_CONFIRMATIONS] = self.__get_min_peer_confirmations()
    return


  def __increment_confirmations(self, key):
    self.__chain_storage[key][self.CS_CONFIRMATIONS] += 1
    return
  
  def __set_confirmations(self, key, confirmations):
    self.__chain_storage[key][self.CS_CONFIRMATIONS] = confirmations
    return


  def __set_key_value(self, key, value, owner, sync_storage=False):
    """
    This method is called to set a key-value pair in the chain storage.
    
    Parameters:
    ----------
    
    key : str
      The key to set the value for
    
    value : any
      The value to set
      
    owner : str
      The owner of the key-value pair
      
    sync_storage : bool
      If `True` will only set the local kv pair without broadcasting to the network. This operation is used for remote sync when
      a node receives a set operation from the network and needs to set the value in the local chain storage replica.
      
    """
    # key should be composed of the chainstore app identity and the actual key
    # so if two chainstore apps are running on the same node, they will not overwrite each other
    # also this way we can implement chaistore app allow-listing
    chain_key = key
    self.__chain_storage[chain_key] = {
      self.CS_KEY   : key,
      self.CS_VALUE : value,
      self.CS_OWNER : owner,
    }    
    self.__reset_confirmations(key)
    if sync_storage:
      # set the confirmations to -1 to indicate that the key is remote synced on this node
      self.__set_confirmations(key, -1) # set to -1 to indicate that the key is remote synced on this node
    self.__save_chain_storage()
    return


  def _set_value(self, key, value, owner=None, debug=False, sync_storage=False):
    """ 
    This method is called to set a value in the chain storage.
    If called locally will push a broadcast request to the network, 
    while if called from the network will set the value in the chain storage.
    
    Parameters:
    ----------
    
    key : str
      The key to set the value for
      
    value : any
      The value to set
      
    owner : str 
      The owner of the key-value pair
      
    debug : bool
      If True will print debug messages
      
    sync_storage : bool
      If True will only set the local kv pair without broadcasting to the network
    
    """
    where = "LOCAL: " if not sync_storage else "REMOTE: "
    debug = debug or self.cfg_chain_store_debug
    if owner is None:
      owner = self.get_instance_path()
    need_store = True
    if key in self.__chain_storage:
      existing_value = self.__get_key_value(key)
      existing_owner = self.__get_key_owner(key)
      if existing_value == value:
        if debug:
          self.P(f" === Key {key} already stored by {existing_owner} has the same value")
        need_store = False
    if need_store:
      if debug:
        self.P(f" === {where}Setting value for key {key}={value} by {owner}, remote={sync_storage}")
      self.__set_key_value(key, value, owner, sync_storage=sync_storage)
      if not sync_storage:      
        # now send set-value confirmation to all
        op = {      
            self.CS_OP : self.CS_STORE,
            self.CS_KEY: key,        
            self.CS_VALUE : value,   
            self.CS_OWNER : owner,
        }
        self.__ops.append(op)
        if debug:
          self.P(f" === {where}Local key {key} locally stored for {owner}. Now waiting for confirmations...")
        # at this point we can wait until we have enough confirmations
        _timeout = self.time() + 10
        _done = False
        _prev_confirm = 0
        _max_retries = 2
        _retries = 0
        while not _done: # this LOCKS the calling thread set_value
          recv_confirm = self.__get_key_confirmations(key)
          if recv_confirm > _prev_confirm:
            _prev_confirm = recv_confirm
            if debug:
              self.P(f" === {where}Key received '{key}' has {recv_confirm} confirmations")
          if recv_confirm >= self.__get_key_min_confirmations(key):
            if debug:
              self.P(f" === {where}KEY CONFIRMED '{key}': has enough ({recv_confirm}) confirmations")
            _done = True
            need_store = True
            continue
          elif self.time() > _timeout:
            if debug:
              self.P(f" === {where}Key '{key}' has not enough confirmations after timeout", color='r')
            _retries += 1
            if _retries > _max_retries:
              if debug:
                self.P(f" === {where}Key '{key}' has not enough confirmations after {_max_retries} retries", color='r')
              _done = True
              need_store = False
            else:
              if debug:
                self.P(f" === {where}Retrying key '{key}' with timeout...", color='r')
              self.__ops.append(op)
              _timeout = self.time() + 10
            # end if retries
          # end if timeout
          self.sleep(0.100)  # sleep for 100ms to give protocol sync time
        # end while not done
      else:
        if debug:
          self.P(f" === {where}key {key} locally stored for {owner} for sync")
      # end if not sync_storage
    # end if need_store
    return need_store


  def _get_value(self, key, get_owner=False, debug=False):
    """ This method is called to get a value from the chain storage """
    debug = debug or self.cfg_chain_store_debug
    if debug:
      self.P(f" === Getting value for key {key}")
    value = self.__get_key_value(key)
    if get_owner:
      owner = self.__get_key_owner(key)
      return value, owner
    return value
  
  ### END setter-getter methods


  def __maybe_broadcast(self):
    """ 
    This method is called to broadcast the chain store operations to the network.
    For each operation in the queue, a broadcast is sent to the network    
    """
    if self.cfg_chain_store_debug and len(self.__ops) > 0:
      self.P(f" === Broadcasting {len(self.__ops)} chain store {self.CS_STORE} ops to {self.__chain_peers}")
    while len(self.__ops) > 0:
      data = {
        self.CS_DATA : self.__ops.popleft()
      }
      self.__send_data_to_chain_peers(data)
    return


  def __exec_store(self, data):
    """ 
    This method is called when a store operation is received from the network. The method will:
      - set the value in the chain storage
      - send a ecrypted confirmation of the storage operation to the network
    """
    key = data.get(self.CS_KEY, None)
    value = data.get(self.CS_VALUE , None)
    owner = data.get(self.CS_OWNER, None)
    if self.cfg_chain_store_debug:
      self.P(f" === REMOTE: Executing remote-sync store for {key}={value} by {owner}")
    result = self._set_value(key, value, owner=owner, sync_storage=True)
    if result:
      # now send confirmation of the storage execution
      if self.cfg_chain_store_debug:
        self.P(f" === REMOTE: Sending storage confirm of {key} by {owner} to {self.__chain_peers}")
      data = {
        self.CS_DATA : {
          self.CS_OP : self.CS_CONFIRM,
          self.CS_KEY: key,
          self.CS_VALUE : value,
          self.CS_OWNER : owner,
          self.CS_CONFIRM_BY : self.get_instance_path(),
          self.CS_CONFIRM_BY_ADDR : self.ee_addr,
        }
      }
      self.__send_data_to_chain_peers(data)
    else:
      if self.cfg_chain_store_debug:
        self.P(f" === REMOTE: Store for {key}={value} of {owner} failed", color='r')
    return


  def __exec_received_confirm(self, data):
    """ This method is called when a confirmation of a broadcasted store operation is received from the network """
    key = data.get(self.CS_KEY, None)
    value = data.get(self.CS_VALUE, None)
    owner = data.get(self.CS_OWNER, None)
    confirm_by = data.get(self.CS_CONFIRM_BY, None)
    op = data.get(self.CS_OP, None)
    
    local_owner = self.__get_key_owner(key)
    local_value = self.__get_key_value(key)
    if self.cfg_chain_store_debug:
      self.P(f" === LOCAL: Received {op} from {confirm_by} for  {key}={value}, owner{owner}")
    if owner == local_owner and value == local_value:
      self.__increment_confirmations(key)
      if self.cfg_chain_store_debug:
        self.P(f" === LOCAL: Key {key} confirmed by {confirm_by}")
    return

  @NetworkProcessorPlugin.payload_handler()
  def default_handler(self, payload):
    sender = payload.get(self.const.PAYLOAD_DATA.EE_SENDER, None)
    destination = payload.get(self.const.PAYLOAD_DATA.EE_DESTINATION, None)
    is_encrypted = payload.get(self.const.PAYLOAD_DATA.EE_IS_ENCRYPTED, False)
    destination = destination if isinstance(destination, list) else [destination]
    decrypted_data = self.receive_and_decrypt_payload(data=payload)    
    # DEBUG AREA
    if self.cfg_full_debug_payloads:
      self.P(f" === on_payload_chain_store_basic from {sender} (enc={is_encrypted})")
      if self.ee_addr in destination:
        self.P(f" === on_payload_chain_store_basic received for me")
      else:
        self.P(f" === on_payload_chain_store_basic to {destination} (not for me)", color='r')
        return
    # try to decrypt the payload
    if self.cfg_full_debug_payloads:
      if decrypted_data is None or len(decrypted_data) == 0:
        self.P(f" === on_payload_chain_store_basic FAILED decrypting payload", color='r')
      else:
        self.P(f" === on_payload_chain_store_basic decrypted payload OK")
    # END DEBUG AREA    
    # get the data and call the appropriate operation method
    data = decrypted_data.get(self.CS_DATA, {})
    operation = data.get(self.CS_OP, None)
    owner = data.get(self.CS_OWNER, None)
    if self.cfg_full_debug_payloads:
      if operation is None:
        self.P(f" === on_payload_chain_store_basic NO OPERATION from data: {data}", color='r')
      else:
        self.P(f" === on_payload_chain_store_basic operation={operation} from owner={owner}")
    if operation == self.CS_STORE:
      self.__exec_store(data)      
    elif operation == self.CS_CONFIRM:
      self.__exec_received_confirm(data)
    return
  

  
  def process(self):
    self.__maybe_refresh_chain_peers()
    self.__maybe_broadcast()
    return 



from naeural_core.serving.mixins_llm.llm_utils import LlmCT


class LlmTokenizerMixin(object):
  def __init__(self, *args, **kwargs):
    super(LlmTokenizerMixin, self).__init__(*args, **kwargs)
    return

  def _set_tokenizer_chat_template(self):
    """
    Update the chat template of the tokenizer for cases
    where transformers doesn't set the correct values.
    For now this covers mistral and llama-3.

    Parameters
    ----------
    None

    Returns
    -------
    None
    """
    if 'mistral' in self.cfg_model_name.lower():
      self.tokenizer.chat_template = LlmCT.MISTRAL_CHAT_TEMPLATE
    if 'llama-3' in self.cfg_model_name.lower():
      self.tokenizer.chat_template = LlmCT.LLAMA3_CHAT_TEMPLATE
    return

  def add_context_to_request(self, request, context):
    """
    Adds context to the request.

    Parameters
    ----------
    request : str
        the request
    context : str
        the context

    Returns
    -------
    str
        the request with context
    """
    return f'{request} - please answer while also considering the following: {context}'

  def _get_prompt_from_template(self, request, history, system_info, context=None):
    """
    Uses Jinja template to generate a prompt.

    Parameters
    ----------
    request : str
        the current request
    history : list[dict]
        the list of previous requests and responses in the same format as for `_get_prompt`
    system_info : str
        the system prompt
    context : str, optional
        the context for the prompt - CURRENTLY DISABLED

    Returns
    -------
    str
        full prompt

    Raises
    ------
    ValueError
        _description_
    """
    chat = []
    if system_info is not None:
      chat.append({LlmCT.ROLE_KEY: LlmCT.SYSTEM_ROLE, LlmCT.DATA_KEY: system_info})

    #endif create system info

    if history is not None and len(history) > 0:
      if not (isinstance(history, list) and isinstance(history[0], dict)):
        msg = "`history` must be a list of dicts. Received {}".format(type(history))
        raise ValueError(msg)
      # endif type check
      if self.cfg_history_limit is not None:
        limit = max(int(self.cfg_history_limit), 0)
        history = history[-limit:]
      # endif history limit
      for chat_round in history:
        round_request = chat_round.get(LlmCT.REQ, None)
        round_response = chat_round.get(LlmCT.RES, None)
        assert round_request is not None, "Each round in `history` must have a `request`"
        assert round_response is not None, "Each round in `history` must have a `response`"
        chat.append({LlmCT.ROLE_KEY: LlmCT.REQUEST_ROLE, LlmCT.DATA_KEY: round_request})
        chat.append({LlmCT.ROLE_KEY: LlmCT.REPLY_ROLE, LlmCT.DATA_KEY: round_response})
      #endfor chat rounds
    #endif history check

    assert isinstance(request, str), "`request` must be a string"
    if False:
      if context is not None and isinstance(context, str):
        request = self.add_context_to_request(request, context)
      # endif context provided
    # The context feature is disabled until further improvements are made.
    chat.append({LlmCT.ROLE_KEY: LlmCT.REQUEST_ROLE, LlmCT.DATA_KEY: request})
    from_template = self.tokenizer.apply_chat_template(
      chat, tokenize=False,
      add_generation_prompt=self.cfg_add_generation_prompt
    )
    return from_template



from naeural_core.business.base import BasePluginExecutor as BasePlugin

__VER__ = '0.1.0.0'

_CONFIG = {
  # mandatory area
  **BasePlugin.CONFIG,

  # our overwritten props
  'AI_ENGINE': "doc_embed",
  "DOC_EMBED_STATUS_PERIOD": 20,
  'ALLOW_EMPTY_INPUTS': True,  # if this is set to true the on-idle will continuously trigger the process

  'VALIDATION_RULES': {
    **BasePlugin.CONFIG['VALIDATION_RULES'],
  },
}


class DocEmbeddingAgentPlugin(BasePlugin):
  CONFIG = _CONFIG

  def on_init(self):
    self.__last_status_time = None
    super(DocEmbeddingAgentPlugin, self).on_init()
    return

  def send_status(self, inf_meta):
    self.add_payload_by_fields(
      contexts=inf_meta.get('contexts', []),
      model_name=inf_meta.get('model_name', ''),
      doc_embed_is_status=True,
    )
    return

  def maybe_send_status(self, inf_meta):
    if self.__last_status_time is None or self.time() - self.__last_status_time > self.cfg_doc_embed_status_period:
      self.send_status(inf_meta)
      self.__last_status_time = self.time()
    # endif
    return

  def _process(self):
    data = self.dataapi_struct_data()
    inf_meta = self.dataapi_inferences_meta().get(self.cfg_ai_engine)
    if inf_meta is None:
      return
    self.maybe_send_status(inf_meta)
    if data is None or len(data) == 0:
      return
    inferences = self.dataapi_struct_data_inferences()
    for inf in inferences:
      # For each inference a response payload will be created
      self.add_payload_by_fields(**inf)
    # endfor inferences
    return
